r"""Loss functions, evaluation metrics, and related utilities."""

import itertools
from typing import Protocol, Optional, Tuple, Union

import math

import torch
from torch import Tensor
from torch.nn.functional import binary_cross_entropy_with_logits, logsigmoid
import torch.nn.functional as F

from deepali.core.enum import FlowDerivativeKeys, SpatialDerivativeKeys
from deepali.core.grid import Grid
from deepali.core.image import avg_pool, dot_channels, rand_sample
from deepali.core.flow import denormalize_flow, divergence, flow_derivatives, spatial_derivatives
from deepali.core.pointset import transform_grid
from deepali.core.pointset import transform_points
from deepali.core.tensor import as_one_hot_tensor, move_dim
from deepali.core.typing import Array, Scalar, ScalarOrTuple


__all__ = (
    "balanced_binary_cross_entropy_with_logits",
    "binary_cross_entropy_with_logits",
    "label_smoothing",
    "dice_score",
    "dice_loss",
    "kld_loss",
    "lcc_loss",
    "mae_loss",
    "mse_loss",
    "ncc_loss",
    "ssd_loss",
    "mi_loss",
    "grad_loss",
    "bending_loss",
    "bending_energy",
    "be_loss",
    "bspline_bending_loss",
    "bspline_bending_energy",
    "bspline_be_loss",
    "curvature_loss",
    "diffusion_loss",
    "divergence_loss",
    "elasticity_loss",
    "focal_loss_with_logits",
    "total_variation_loss",
    "tv_loss",
    "tversky_index",
    "tversky_index_with_logits",
    "tversky_loss",
    "tversky_loss_with_logits",
    "inverse_consistency_loss",
    "masked_loss",
    "reduce_loss",
    "wlcc_loss",
)


class ElementwiseLoss(Protocol):
    r"""Type annotation of a eleemntwise loss function."""

    def __call__(self, input: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
        ...


def label_smoothing(
    labels: Tensor,
    num_classes: Optional[int] = None,
    ignore_index: Optional[int] = None,
    alpha: float = 0.1,
) -> Tensor:
    r"""Apply label smoothing to target labels.

    Implements label smoothing as proposed by Muller et al., (2019) in https://arxiv.org/abs/1906.02629v2.

    Args:
        labels: Scalar target labels or one-hot encoded class probabilities.
        num_classes: Number of target labels. If ``None``, use maximum value in ``target`` plus 1
            when a scalar label map is given.
        ignore_index: Ignore index to be kept during the expansion. The locations of the index
            value in the labels image is stored in the corresponding locations across all channels
            so that this location can be ignored across all channels later, e.g. in Dice computation.
            This argument must be ``None`` if ``labels`` has ``C == num_channels``.
        alpha: Label smoothing factor in [0, 1]. If zero, no label smoothing is done.

    Returns:
        Multi-channel tensor of smoothed target class probabilities.

    """
    if not isinstance(labels, Tensor):
        raise TypeError("label_smoothing() 'labels' must be Tensor")
    if labels.ndim < 4:
        raise ValueError("label_smoothing() 'labels' must be tensor of shape (N, C, ..., X)")
    if labels.shape[1] == 1:
        target = as_one_hot_tensor(
            labels, num_classes, ignore_index=ignore_index, dtype=torch.float32
        )
    else:
        target = labels.float()
    if alpha > 0:
        target = (1 - alpha) * target + alpha * (1 - target) / (target.size(1) - 1)
    return target


def balanced_binary_cross_entropy_with_logits(
    logits: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Balanced binary cross entropy.

    Shruti Jadon (2020) A survey of loss functions for semantic segmentation.
    https://arxiv.org/abs/2006.14822

    Args:
        logits: Logits of binary predictions as tensor of shape ``(N, 1, ..., X)``.
        target: Target label probabilities as tensor of shape ``(N, 1, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, 1, ..., X)``.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        Balanced binary cross entropy (bBCE). If ``reduction="none"``, the returned tensor has shape
        ``(N, 1, ..., X)`` with bBCE for each element. Otherwise, it is reduced into a scalar.

    """
    if logits.ndim < 3 or logits.shape[1] != 1:
        raise ValueError(
            "balanced_binary_cross_entropy_with_logits() 'logits' must have shape (N, 1, ..., X)"
        )
    if target.ndim < 3 or target.shape[1] != 1:
        raise ValueError(
            "balanced_binary_cross_entropy_with_logits() 'target' must have shape (N, 1, ..., X)"
        )
    if logits.shape[0] != target.shape[0]:
        raise ValueError(
            "balanced_binary_cross_entropy_with_logits() 'logits' and 'target' must have matching batch size"
        )
    neg_weight = target.flatten(1).mean(-1).reshape((-1,) + (1,) * (target.ndim - 1))
    pos_weight = 1 - neg_weight
    log_y_pred: Tensor = logsigmoid(logits)
    loss_pos = -log_y_pred.mul(target)
    loss_neg = logits.sub(log_y_pred).mul(1 - target)
    loss = loss_pos.mul(pos_weight).add(loss_neg.mul(neg_weight))
    loss = masked_loss(loss, weight, "balanced_binary_cross_entropy_with_logits", inplace=True)
    loss = reduce_loss(loss, reduction=reduction)
    return loss


def dice_score(
    input: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    epsilon: float = 1e-15,
    reduction: str = "mean",
) -> Tensor:
    r"""Soft Dice similarity of multi-channel classification result.

    Args:
        input: Normalized logits of binary predictions as tensor of shape ``(N, C, ..., X)``.
        target: Target label probabilities as tensor of shape ``(N, C, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, C, ..., X)``.
        epsilon: Small constant used to avoid division by zero.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        Dice similarity coefficient (DSC). If ``reduction="none"``, the returned tensor has shape
        ``(N, C)`` with DSC for each batch. Otherwise, the DSC scores are reduced into a scalar.

    """
    if not isinstance(input, Tensor):
        raise TypeError("dice_score() 'input' must be torch.Tensor")
    if not isinstance(target, Tensor):
        raise TypeError("dice_score() 'target' must be torch.Tensor")
    if input.dim() < 3:
        raise ValueError("dice_score() 'input' must be tensor of shape (N, C, ..., X)")
    if input.shape != target.shape:
        raise ValueError("dice_score() 'input' and 'target' must have identical shape")
    y_pred = input.float()
    y = target.float()
    intersection = dot_channels(y_pred, y, weight=weight)
    denominator = dot_channels(y_pred, y_pred, weight=weight) + dot_channels(y, y, weight=weight)
    loss = intersection.mul_(2).add_(epsilon).div(denominator.add_(epsilon))
    loss = reduce_loss(loss, reduction)
    return loss


def dice_loss(
    input: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    epsilon: float = 1e-15,
    reduction: str = "mean",
) -> Tensor:
    r"""One minus soft Dice similarity of multi-channel classification result.

    Args:
        input: Normalized logits of binary predictions as tensor of shape ``(N, C, ..., X)``.
        target: Target label probabilities as tensor of shape ``(N, C, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, C, ..., X)``.
        epsilon: Small constant used to avoid division by zero.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        One minus Dice similarity coefficient (DSC). If ``reduction="none"``, the returned tensor has shape
        ``(N, C)`` with DSC for each batch. Otherwise, the DSC scores are reduced into a scalar.

    """
    dsc = dice_score(input, target, weight=weight, epsilon=epsilon, reduction="none")
    loss = reduce_loss(1 - dsc, reduction)
    return loss


def tversky_index(
    input: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    alpha: Optional[float] = None,
    beta: Optional[float] = None,
    epsilon: float = 1e-15,
    normalize: bool = False,
    binarize: bool = False,
    reduction: str = "mean",
) -> Tensor:
    r"""Tversky index as described in https://arxiv.org/abs/1706.05721.

    Args:
        input: Binary predictions as tensor of shape ``(N, 1, ..., X)``
            or multi-class prediction tensor of shape ``(N, C, ..., X)``.
        target: Target labels as tensor of shape ``(N, ..., X)``, binary classification target
            of shape ``(N, 1, ..., X)``, or one-hot encoded tensor of shape ``(N, C, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, ..., X)`` or ``(N, 1|C, ..., X)``..
        alpha: False positives multiplier. Set to ``1 - beta`` if ``None``.
        beta: False negatives multiplier.
        epsilon: Constant used to avoid division by zero.
        normalize: Whether to normalize ``input`` predictions using ``sigmoid`` or ``softmax``.
        binarize: Whether to round normalized predictions to 0 or 1, respectively. If ``False``,
            soft normalized predictions (and target scores) are used. In order for the Tversky
            index to be identical to Dice, this option must be set to ``True`` and ``alpha=beta=0.5``.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        Tversky index (TI). If ``reduction="none"``, the returned tensor has shape ``(N, C)``
        with TI for each batch. Otherwise, the TI values are reduced into a scalar.

    """
    if alpha is None and beta is None:
        alpha = beta = 0.5
    elif alpha is None:
        alpha = 1 - beta
    elif beta is None:
        beta = 1 - alpha
    if not isinstance(input, Tensor):
        raise TypeError("tversky_index() 'input' must be torch.Tensor")
    if not isinstance(target, Tensor):
        raise TypeError("tversky_index() 'target' must be torch.Tensor")
    if input.ndim < 3 or input.shape[1] < 1:
        raise ValueError(
            "tversky_index() 'input' must be have shape (N, 1, ..., X) or (N, C>1, ..., X)"
        )
    if target.ndim < 2 or target.shape[1] < 1:
        raise ValueError(
            "tversky_index() 'target' must be have shape (N, ..., X), (N, 1, ..., X), or (N, C>1, ..., X)"
        )
    if target.shape[0] != input.shape[0]:
        raise ValueError(
            "tversky_index() 'input' and 'target' batch size must be identical"
            f", got {input.shape[0]} != {target.shape[0]}"
        )
    input = input.float()
    if input.shape[1] == 1:
        y_pred = input.sigmoid() if normalize else input
    else:
        y_pred = input.softmax(1) if normalize else input
    if binarize:
        y_pred = y_pred.round()
    num_classes = max(2, y_pred.shape[1])
    if target.ndim == input.ndim:
        y = target.type(y_pred.dtype)
        if target.shape[1] == 1:
            if num_classes > 2:
                raise ValueError(
                    "tversky_index() 'target' has shape (N, 1, ..., X)"
                    f", but 'input' is multi-class prediction (C={num_classes})"
                )
            if y_pred.shape[1] == 2:
                y_pred = y_pred.narrow(1, 1, 1)
        else:
            if target.shape[1] != num_classes:
                raise ValueError(
                    "tversky_index() 'target' has shape (N, C, ..., X)"
                    f", but C does not match 'input' with C={num_classes}"
                )
            if y_pred.shape[1] == 1:
                y = y.narrow(1, 1, 1)
        if binarize:
            y = y.round()
    elif target.ndim + 1 == y_pred.ndim:
        if num_classes == 2 and y_pred.shape[1] == 1:
            y = target.unsqueeze(1).ge(0.5).type(y_pred.dtype)
            if binarize:
                y = y.round()
        else:
            y = as_one_hot_tensor(target, num_classes, dtype=y_pred.dtype)
    else:
        raise ValueError(
            "tversky_index() 'target' must be tensor of shape (N, ..., X) or (N, C, ... X)"
        )
    if y.shape != y_pred.shape:
        raise ValueError("tversky_index() 'input' and 'target' shapes must be compatible")
    if weight is not None:
        if weight.ndim + 1 == y.ndim:
            weight = weight.unsqueeze(1)
        if weight.ndim != y.ndim:
            raise ValueError("tversky_index() 'weight' shape must be (N, ..., X) or (N, C, ..., X)")
        if weight.shape[0] != target.shape[0]:
            raise ValueError(
                "tversky_index() 'weight' batch size does not match 'input' and 'target'"
            )
        if weight.shape[1] == 1:
            weight = weight.repeat((1,) + (num_classes,) + (1,) * (weight.ndim - 2))
        if weight.shape != y.shape:
            raise ValueError(
                "tversky_index() 'weight' shape must be compatible with 'input' and 'target'"
            )
    intersection = dot_channels(y_pred, y, weight=weight)
    fps = dot_channels(y_pred, 1 - y, weight=weight).mul_(alpha)
    fns = dot_channels(1 - y_pred, y, weight=weight).mul_(beta)
    numerator = intersection.add_(epsilon)
    denominator = numerator.add(fps).add(fns)
    loss = numerator.div(denominator)
    loss = reduce_loss(loss, reduction)
    return loss


def tversky_index_with_logits(
    logits: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    alpha: Optional[float] = None,
    beta: Optional[float] = None,
    epsilon: float = 1e-15,
    binarize: bool = False,
    reduction: str = "mean",
) -> Tensor:
    r"""Tversky index as described in https://arxiv.org/abs/1706.05721.

    Args:
        logits: Binary prediction logits as tensor of shape ``(N, 1, ..., X)``.
        target: Target labels as tensor of shape ``(N, ..., X)`` or ``(N, 1, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, ..., X)`` or ``(N, 1, ..., X)``.
        alpha: False positives multiplier. Set to ``1 - beta`` if ``None``.
        beta: False negatives multiplier.
        epsilon: Constant used to avoid division by zero.
        normalize: Whether to normalize ``input`` predictions using ``sigmoid`` or ``softmax``.
        binarize: Whether to round normalized predictions to 0 or 1, respectively. If ``False``,
            soft normalized predictions (and target scores) are used. In order for the Tversky
            index to be identical to Dice, this option must be set to ``True`` and ``alpha=beta=0.5``.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        Tversky index (TI). If ``reduction="none"``, the returned tensor has shape ``(N, 1)``
        with TI for each batch. Otherwise, the TI values are reduced into a scalar.

    """
    return tversky_index(
        logits,
        target,
        weight=weight,
        alpha=alpha,
        beta=beta,
        epsilon=epsilon,
        normalize=True,
        binarize=binarize,
        reduction=reduction,
    )


def tversky_loss(
    input: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    alpha: Optional[float] = None,
    beta: Optional[float] = None,
    gamma: Optional[float] = None,
    epsilon: float = 1e-15,
    normalize: bool = False,
    binarize: bool = False,
    reduction: str = "mean",
) -> Tensor:
    r"""Tversky loss as described in https://arxiv.org/abs/1706.05721.

    Args:
        input: Normalized logits of binary predictions as tensor of shape ``(N, C, ..., X)``.
        target: Target label probabilities as tensor of shape ``(N, C, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, ..., X)`` or ``(N, 1, ..., X)``.
        alpha: False positives multiplier. Set to ``1 - beta`` if ``None``.
        beta: False negatives multiplier.
        gamma: Exponent used for focal Tverksy loss.
        epsilon: Constant used to avoid division by zero.
        normalize: Whether to normalize ``input`` predictions using ``sigmoid`` or ``softmax``.
        binarize: Whether to round normalized predictions to 0 or 1, respectively. If ``False``,
            soft normalized predictions (and target scores) are used. In order for the Tversky
            index to be identical to Dice, this option must be set to ``True`` and ``alpha=beta=0.5``.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        One minus Tversky index (TI) to the power of gamma. If ``reduction="none"``, the returned
        tensor has shape ``(N, C)`` with the loss for each batch. Otherwise, a scalar is returned.

    """
    ti = tversky_index(
        input,
        target,
        weight=weight,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        epsilon=epsilon,
        normalize=normalize,
        binarize=binarize,
        reduction="none",
    )
    one = torch.tensor(1, dtype=ti.dtype, device=ti.device)
    loss = one.sub(ti)
    if gamma:
        if gamma > 1:
            loss = loss.pow_(gamma)
        elif gamma < 1:
            raise ValueError("tversky_loss() 'gamma' must be greater than or equal to 1")
    loss = reduce_loss(loss, reduction)
    return loss


def tversky_loss_with_logits(
    logits: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    alpha: Optional[float] = None,
    beta: Optional[float] = None,
    gamma: Optional[float] = None,
    epsilon: float = 1e-15,
    binarize: bool = False,
    reduction: str = "mean",
) -> Tensor:
    r"""Tversky loss as described in https://arxiv.org/abs/1706.05721.

    Args:
        logits: Binary prediction logits as tensor of shape ``(N, 1, ..., X)``.
        target: Target labels as tensor of shape ``(N, ..., X)`` or ``(N, 1, ..., X)``.
        weight: Voxelwise label weight tensor of shape ``(N, ..., X)`` or ``(N, 1, ..., X)``.
        alpha: False positives multiplier. Set to ``1 - beta`` if ``None``.
        beta: False negatives multiplier.
        gamma: Exponent used for focal Tverksy loss.
        epsilon: Constant used to avoid division by zero.
        normalize: Whether to normalize ``input`` predictions using ``sigmoid`` or ``softmax``.
        binarize: Whether to round normalized predictions to 0 or 1, respectively. If ``False``,
            soft normalized predictions (and target scores) are used. In order for the Tversky
            index to be identical to Dice, this option must be set to ``True`` and ``alpha=beta=0.5``.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        One minus Tversky index (TI) to the power of gamma. If ``reduction="none"``, the returned
        tensor has shape ``(N, C)`` with the loss for each batch. Otherwise, a scalar is returned.

    """
    return tversky_loss(
        logits,
        target,
        weight=weight,
        alpha=alpha,
        beta=beta,
        gamma=gamma,
        epsilon=epsilon,
        normalize=True,
        binarize=binarize,
        reduction=reduction,
    )


def focal_loss_with_logits(
    input: Tensor,
    target: Tensor,
    weight: Optional[Tensor] = None,
    alpha: float = 0.25,
    gamma: float = 2,
    reduction: str = "mean",
) -> Tensor:
    """Loss used in RetinaNet for dense detection: https://arxiv.org/abs/1708.02002.

    Args:
        input: Logits of the predictions for each example.
        target: A tensor with the same shape as ``input``. Stores the binary classification
            label for each element in inputs (0 for the negative class and 1 for the positive class).
        weight: Multiplicative mask tensor with same shape as ``input``.
        alpha: Weighting factor in [0, 1] to balance positive vs negative examples or -1 for ignore.
        gamma: Exponent of the modulating factor (1 - p_t) to balance easy vs hard examples.
        reduction: Either ``none``, ``mean``, or ``sum``.

    Returns:
        Loss tensor with the reduction option applied.

    """
    # https://www.kaggle.com/bigironsphere/loss-function-library-keras-pytorch#Focal-Loss
    bce = binary_cross_entropy_with_logits(input, target, reduction="none")
    one = torch.tensor(1, dtype=bce.dtype, device=bce.device)
    loss = one.sub(torch.exp(-bce)).pow(gamma).mul(bce)
    if alpha >= 0:
        if alpha > 1:
            raise ValueError("focal_loss() 'alpha' must be in [0, 1]")
        loss = target.mul(alpha).add(one.sub(target).mul(1 - alpha)).mul(loss)
    loss = masked_loss(loss, weight, "focal_loss_with_logits", inplace=True)
    loss = reduce_loss(loss, reduction)
    return loss


def kld_loss(mean: Tensor, logvar: Tensor, reduction: str = "mean") -> Tensor:
    r"""Kullback-Leibler divergence in case of zero-mean and isotropic unit variance normal prior distribution.

    Kingma and Welling, Auto-Encoding Variational Bayes, ICLR 2014, https://arxiv.org/abs/1312.6114 (Appendix B).

    """
    loss = mean.square().add_(logvar.exp()).sub_(1).sub_(logvar)
    loss = reduce_loss(loss, reduction)
    loss = loss.mul_(0.5)
    return loss


def ncc_loss(
    source: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    epsilon: float = 1e-15,
    reduction: str = "mean",
) -> Tensor:
    r"""Normalized cross correlation.

    Args:
        source: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``source``.
        mask: Multiplicative mask tensor with same shape as ``source``.
        epsilon: Small constant added to denominator to avoid division by zero.
        reduction: Whether to compute "mean" or "sum" of normalized cross correlation
            of image pairs in batch. If "none", a 1-dimensional tensor is returned
            with length equal the batch size.

    Returns:
        Negative squared normalized cross correlation plus one.

    """

    if not isinstance(source, Tensor):
        raise TypeError("ncc_loss() 'source' must be tensor")
    if not isinstance(target, Tensor):
        raise TypeError("ncc_loss() 'target' must be tensor")
    if source.shape != target.shape:
        raise ValueError("ncc_loss() 'source' must have same shape as 'target'")

    source = source.reshape(source.shape[0], -1).float()
    target = target.reshape(source.shape[0], -1).float()

    source_mean = source.mean(dim=1, keepdim=True)
    target_mean = target.mean(dim=1, keepdim=True)

    x = source.sub(source_mean)
    y = target.sub(target_mean)

    a = x.mul(y).sum(dim=1)
    b = x.square().sum(dim=1)
    c = y.square().sum(dim=1)

    loss = a.square_().div_(b.mul_(c).add_(epsilon)).neg_().add_(1)
    loss = masked_loss(loss, mask, "ncc_loss")
    loss = reduce_loss(loss, reduction, mask)
    return loss


def lcc_loss(
    source: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    kernel_size: ScalarOrTuple[int] = 7,
    epsilon: float = 1e-15,
    reduction: str = "mean",
) -> Tensor:
    r"""Local normalized cross correlation.

    References:
        Avants et al., 2008, Symmetric Diffeomorphic Image Registration with Cross Correlation:
            Evaluating Automated Labeling of Elderly and Neurodegenerative Brain,
            doi:10.1016/j.media.2007.06.004.

    Args:
        source: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``source``.
        mask: Multiplicative mask tensor with same shape as ``source``.
        kernel_size: Local rectangular window size in number of grid points.
        epsilon: Small constant added to denominator to avoid division by zero.
        reduction: Whether to compute "mean" or "sum" over all grid points. If "none", output
            tensor shape is equal to the shape of the input tensors given an odd kernel size.

    Returns:
        Negative local normalized cross correlation plus one.

    """

    if not isinstance(source, Tensor):
        raise TypeError("lcc_loss() 'source' must be tensor")
    if not isinstance(target, Tensor):
        raise TypeError("lcc_loss() 'target' must be tensor")
    if source.shape != target.shape:
        raise ValueError("lcc_loss() 'source' must have same shape as 'target'")

    def local_sum(data: Tensor) -> Tensor:
        return avg_pool(
            data,
            kernel_size=kernel_size,
            stride=1,
            padding=None,
            divisor_override=1,
        )

    def local_mean(data: Tensor) -> Tensor:
        return avg_pool(
            data,
            kernel_size=kernel_size,
            stride=1,
            padding=None,
            count_include_pad=False,
        )

    source = source.float()
    target = target.float()

    source_mean = local_mean(source)
    target_mean = local_mean(target)

    x = source.sub(source_mean)
    y = target.sub(target_mean)

    a = local_sum(x.mul(y))
    b = local_sum(x.square())
    c = local_sum(y.square())

    loss = a.square_().div_(b.mul_(c).add_(epsilon)).neg_().add_(1)
    loss = masked_loss(loss, mask, "lcc_loss")
    loss = reduce_loss(loss, reduction, mask)
    return loss


def wlcc_loss(
    source: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    source_mask: Optional[Tensor] = None,
    target_mask: Optional[Tensor] = None,
    kernel_size: ScalarOrTuple[int] = 7,
    epsilon: float = 1e-15,
    reduction: str = "mean",
) -> Tensor:
    r"""Weighted local normalized cross correlation.

    References:
        Lewis et al., 2020, Fast Learning-based Registration of Sparse 3D Clinical Images, arXiv:1812.06932.

    Args:
        source: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``source``.
        mask: Multiplicative mask tensor ``w_c`` with same shape as ``target`` and ``source``.
            This tensor is used for computing the weighted local correlation. If ``None`` and
            both ``source_mask`` and ``target_mask`` are given, it is set to the product of these.
            Otherwise, no mask is used to aggregate the local cross correlation values. When both
            ``source_mask`` and ``target_mask`` are ``None``, but ``mask`` is not, then the specified
            ``mask`` is used both as ``source_mask`` and ``target_mask``.
        source_mask: Multiplicative mask tensor ``w_m`` with same shape as ``source``.
            This tensor is used for computing the weighted local ``source`` mean. If ``None``,
            the local mean is computed over all ``source`` elements within each local neighborhood.
        target_mask: Multiplicative mask tensor ``w_f`` with same shape as ``source``.
            This tensor is used for computing the weighted local ``target`` mean. If ``None``,
            the local mean is computed over all ``target`` elements within each local neighborhood.
        kernel_size: Local rectangular window size in number of grid points.
        epsilon: Small constant added to denominator to avoid division by zero.
        reduction: Whether to compute "mean" or "sum" over all grid points. If "none", output
            tensor shape is equal to the shape of the input tensors given an odd kernel size.

    Returns:
        Negative local normalized cross correlation plus one.

    """

    if not isinstance(source, Tensor):
        raise TypeError("wlcc_loss() 'source' must be tensor")
    if not isinstance(target, Tensor):
        raise TypeError("wlcc_loss() 'target' must be tensor")

    if source.shape != target.shape:
        raise ValueError("wlcc_loss() 'source' must have same shape as 'target'")

    for t, t_name, w, w_name in zip(
        [target, source, target],
        ["target", "source", "target"],
        [mask, source_mask, target_mask],
        ["mask", "source_mask", "target_mask"],
    ):
        if w is None:
            continue
        if not isinstance(w, Tensor):
            raise TypeError(f"wlcc_loss() '{w_name}' must be tensor")
        if w.shape[0] not in (1, t.shape[0]):
            raise ValueError(
                f"wlcc_loss() '{w_name}' batch size ({w.shape[0]}) must be 1 or match '{t_name}' ({t.shape[0]})"
            )
        if w.shape[1] not in (1, t.shape[1]):
            raise ValueError(
                f"wlcc_loss() '{w_name}' number of channels ({w.shape[1]}) must be 1 or match '{t_name}' ({t.shape[1]})"
            )
        if w.shape[2:] != t.shape[2:]:
            raise ValueError(
                f"wlcc_loss() '{w_name}' grid shape ({w.shape[2:]}) must match '{t_name}' ({t.shape[2:]})"
            )

    def local_sum(data: Tensor) -> Tensor:
        return avg_pool(
            data,
            kernel_size=kernel_size,
            stride=1,
            padding=None,
            divisor_override=1,
        )

    def local_mean(data: Tensor, weight: Optional[Tensor] = None) -> Tensor:
        if weight is None:
            return avg_pool(
                data,
                kernel_size=kernel_size,
                stride=1,
                padding=None,
                count_include_pad=False,
            )
        a = local_sum(data.mul(weight))
        b = local_sum(weight).add_(epsilon)
        return a.div_(b)

    if mask is not None and source_mask is None and target_mask is None:
        source_mask = mask.float()
        target_mask = source_mask
    else:
        if source_mask is not None:
            source_mask = source_mask.float()
        if target_mask is not None:
            target_mask = target_mask.float()

    source = source.float()
    target = target.float()

    source_mean = local_mean(source, source_mask)
    target_mean = local_mean(target, target_mask)

    x = source.sub(source_mean)
    y = target.sub(target_mean)

    if mask is None and source_mask is not None and target_mask is not None:
        mask = source_mask.mul(target_mask)
    if mask is not None:
        x = x.mul_(mask)
        y = y.mul_(mask)

    a = local_sum(x.mul(y))
    b = local_sum(x.square())
    c = local_sum(y.square())

    loss = a.square_().div_(b.mul_(c).add_(epsilon)).neg_().add_(1)
    loss = masked_loss(loss, mask, name="wlcc_loss")
    loss = reduce_loss(loss, reduction, mask)
    return loss


def huber_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "mean",
    delta: float = 1.0,
) -> Tensor:
    r"""Normalized masked Huber loss.

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.
        delta: Specifies the threshold at which to change between delta-scaled L1 and L2 loss.

    Returns:
        Masked, aggregated, and normalized Huber loss.

    """

    def loss_fn(input: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
        return F.huber_loss(input, target, reduction=reduction, delta=delta)

    return elementwise_loss(
        "huber_loss", loss_fn, input, target, mask=mask, norm=norm, reduction=reduction
    )


def smooth_l1_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "mean",
    beta: float = 1.0,
) -> Tensor:
    r"""Normalized masked smooth L1 loss.

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.
        delta: Specifies the threshold at which to change between delta-scaled L1 and L2 loss.

    Returns:
        Masked, aggregated, and normalized smooth L1 loss.

    """

    def loss_fn(input: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
        return F.smooth_l1_loss(input, target, reduction=reduction, beta=beta)

    return elementwise_loss(
        "smooth_l1_loss", loss_fn, input, target, mask=mask, norm=norm, reduction=reduction
    )


def l1_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Normalized mean absolute error.

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.

    Returns:
        Normalized mean absolute error.

    """

    def loss_fn(input: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
        return F.l1_loss(input, target, reduction=reduction)

    return elementwise_loss(
        "l1_loss", loss_fn, input, target, mask=mask, norm=norm, reduction=reduction
    )


def mae_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Normalized mean absolute error.

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.

    Returns:
        Normalized mean absolute error.

    """

    def loss_fn(input: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
        return F.l1_loss(input, target, reduction=reduction)

    return elementwise_loss(
        "mae_loss", loss_fn, input, target, mask=mask, norm=norm, reduction=reduction
    )


def mse_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Average normalized squared differences.

    This loss is equivalent to `ssd_loss`, except that the default `reduction` is "mean".

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.

    Returns:
        Average normalized squared differences.

    """
    return ssd_loss(input, target, mask=mask, norm=norm, reduction=reduction)


def ssd_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "sum",
) -> Tensor:
    r"""Sum of normalized squared differences.

    The SSD loss is equivalent to MSE, except that an optional overlap mask is supported and
    that the loss value is optionally multiplied by a normalization constant. Moreover, by default
    the sum instead of the mean of per-element loss values is returned (cf. ``reduction``).
    The value returned by ``max_difference(source, target).square()`` can be used as normalization
    factor, which is equvalent to first normalizing the images to [0, 1].

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.

    Returns:
        Sum of normalized squared differences.

    """
    if not isinstance(input, Tensor):
        raise TypeError("ssd_loss() 'input' must be tensor")
    if not isinstance(target, Tensor):
        raise TypeError("ssd_loss() 'target' must be tensor")
    if input.shape != target.shape:
        raise ValueError("ssd_loss() 'input' must have same shape as 'target'")
    loss = input.sub(target).square()
    loss = masked_loss(loss, mask, "ssd_loss")
    loss = reduce_loss(loss, reduction, mask)
    if norm is not None:
        norm = torch.as_tensor(norm, dtype=loss.dtype, device=loss.device).squeeze()
        if not norm.ndim == 0:
            raise ValueError("ssd_loss() 'norm' must be scalar")
        if norm > 0:
            loss = loss.div_(norm)
    return loss


def mi_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    num_bins: Optional[int] = None,
    num_samples: Optional[int] = None,
    sample_ratio: Optional[float] = None,
    normalized: bool = False,
) -> Tensor:
    r"""Calculate mutual information loss using Parzen window density and entropy estimations.

    References:
        Qiu, H., Qin, C., Schuh, A., Hammernik, K.: Learning Diffeomorphic and Modality-invariant
            Registration using B-splines. Medical Imaging with Deep Learning. (2021).
        Thévenaz, P., Unser, M.: Optimization of mutual information for multiresolution image registration.
            IEEE Trans. Image Process. 9, 2083–2099 (2000).

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Region of interest mask with same shape as ``input``.
        vmin: Minimal intensity value the joint and marginal density is estimated.
        vmax: Maximal intensity value the joint and marginal density is estimated.
        num_bins: Number of bin edges to discretize the density estimation.
        num_samples: Number of voxels in the image domain randomly sampled to compute the loss,
            ignored if `sample_ratio` is also set.
        sample_ratio: Ratio of voxels in the image domain randomly sampled to compute the loss.
        normalized: Calculate Normalized Mutual Information instead of Mutual Information if True.

    Returns:
        Negative mutual information. If ``normalized=True``, 2 is added such that the loss value is in [0, 2].

    """

    if target.ndim < 3:
        raise ValueError("mi_loss() 'target' must be tensor of shape (N, C, ..., X)")
    if input.shape != target.shape:
        raise ValueError("ssd_loss() 'input' must have same shape as 'target'")

    if vmin is None:
        vmin = torch.min(input.min(), target.min()).item()
    if vmax is None:
        vmax = torch.max(input.max(), target.max()).item()
    if num_bins is None:
        num_bins = 64
    elif num_bins == "auto":
        raise NotImplementedError(
            "mi_loss() automatically setting num_bins based on dynamic range of input"
        )

    # Flatten spatial dimensions of inputs
    shape = target.shape
    input = input.flatten(2)
    target = target.flatten(2)

    if mask is not None:
        if mask.ndim < 3 or mask.shape[2:] != shape[2:] or mask.shape[1] != 1:
            raise ValueError(
                "mi_loss() 'mask' must be tensor of shape (1|N, 1, ..., X) with spatial dimensions matching 'target'"
            )
        mask = mask.flatten(2)

    # Random image samples, optionally weighted by mask
    if sample_ratio is not None:
        if sample_ratio <= 0 or sample_ratio > 1:
            raise ValueError("mi_loss() 'sample_ratio' must be in open-closed interval (0, 1]")
        if num_samples is None:
            num_samples = target.shape[2:].numel()
        num_samples = min(max(1, int(sample_ratio * target.shape[2:].numel())), num_samples)
    if num_samples is not None:
        input, target = rand_sample([input, target], num_samples, mask=mask, replacement=True)
    elif mask is not None:
        input = input.mul(mask)
        target = target.mul(mask)

    # set the bin edges and Gaussian kernel std
    bin_width = (vmax - vmin) / num_bins  # FWHM is one bin width
    bin_center = torch.linspace(vmin, vmax, num_bins, requires_grad=False)
    bin_center = bin_center.unsqueeze(1).type_as(input)

    # calculate Parzen window function response
    pw_sdev = bin_width * (1 / (2 * math.sqrt(2 * math.log(2))))
    pw_norm = 1 / math.sqrt(2 * math.pi) * pw_sdev

    def parzen_window_fn(x: Tensor) -> Tensor:
        return x.sub(bin_center).square().div(2 * pw_sdev**2).neg().exp().mul(pw_norm)

    pw_input = parzen_window_fn(input)  # (N, #bins, H*W*D)
    pw_target = parzen_window_fn(target)

    # calculate joint histogram
    hist_joint = pw_input.bmm(pw_target.transpose(1, 2))  # (N, #bins, #bins)
    hist_norm = hist_joint.flatten(start_dim=1, end_dim=-1).sum(dim=1).add_(1e-5)

    # joint and marginal distributions
    p_joint = hist_joint / hist_norm.view(-1, 1, 1)  # (N, #bins, #bins) / (N, 1, 1)
    p_input = p_joint.sum(dim=2)
    p_target = p_joint.sum(dim=1)

    # calculate entropy
    ent_input = -torch.sum(p_input * torch.log(p_input + 1e-5), dim=1)  # (N,1)
    ent_target = -torch.sum(p_target * torch.log(p_target + 1e-5), dim=1)  # (N,1)
    ent_joint = -torch.sum(p_joint * torch.log(p_joint + 1e-5), dim=(1, 2))  # (N,1)

    if normalized:
        loss = 2 - torch.mean((ent_input + ent_target) / ent_joint)
    else:
        loss = torch.mean(ent_input + ent_target - ent_joint).neg()
    return loss


def nmi_loss(
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    vmin: Optional[float] = None,
    vmax: Optional[float] = None,
    num_bins: Optional[int] = None,
    num_samples: Optional[int] = None,
    sample_ratio: Optional[float] = None,
) -> Tensor:
    return mi_loss(
        input,
        target,
        mask=mask,
        vmin=vmin,
        vmax=vmax,
        num_bins=num_bins,
        num_samples=num_samples,
        sample_ratio=sample_ratio,
        normalized=True,
    )


def grad_loss(
    u: Tensor,
    p: Union[int, float] = 2,
    q: Optional[Union[int, float]] = 1,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Loss term based on p-norm of spatial gradient of vector fields.

    The ``p`` and ``q`` parameters can be used to specify which norm to compute, i.e., ``sum(abs(du)**p)**q``,
    where ``du`` are the 1st order spatial derivative of the input vector fields ``u`` computed using a finite
    difference scheme and optionally normalized using a specified grid ``spacing``.

    This regularization loss is the basis, for example, for total variation and diffusion penalties.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        p: The order of the gradient norm. When ``p = 0``, the partial derivatives are summed up.
        q: Power parameter of gradient norm. If ``None``, then ``q = 1 / p``. If ``q == 0``, the
            absolute value of the sum of partial derivatives is computed at each grid point.
        mode: Method used to approximate :func:`spatial_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Spatial gradient loss of vector fields.

    """
    if u.ndim < 4:
        # No loss for homogeneous coordinate transformations
        if reduction == "none":
            raise NotImplementedError(
                "grad_loss() not implemented for linear transformation and 'reduction'='none'"
            )
        return torch.tensor(0, dtype=u.dtype, device=u.device)
    N = u.shape[0]
    D = u.shape[1]
    if u.ndim - 2 != D:
        raise ValueError("grad_loss() 'u' must be tensor of shape (N, D, ..., X)")
    if q is None:
        q = 1.0 / p
    if spacing is None:
        spacing = tuple(reversed([2 / (n - 1) for n in u.shape[2:]]))
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    which = SpatialDerivativeKeys.all(spatial_dims=D, order=1)
    deriv = spatial_derivatives(u, which=which, **kwargs)
    if p == 1:
        deriv = {k: v.abs_() for k, v in deriv.items()}
    elif p != 0:
        if p % 2 == 0:
            deriv = {k: v.pow_(p) for k, v in deriv.items()}
        else:
            deriv = {k: v.abs_().pow_(p) for k, v in deriv.items()}
    loss = torch.zeros((N, 1) + u.shape[2:], dtype=u.dtype, device=u.device)
    for value in deriv.values():
        loss = loss.add_(value.sum(dim=1, keepdim=True))
    if q == 0:
        loss.abs_()
    elif q != 1:
        loss.pow_(q)
    loss = reduce_loss(loss, reduction)
    return loss


def bending_loss(
    u: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Bending energy of vector fields.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        mode: Method used to approximate :func:`flow_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Bending energy.

    """
    if u.ndim < 4:
        # No loss for homogeneous coordinate transformations
        if reduction == "none":
            raise NotImplementedError(
                "bending_energy() not implemented for linear transformation and 'reduction'='none'"
            )
        return torch.tensor(0, dtype=u.dtype, device=u.device)
    D = u.shape[1]
    if u.ndim - 2 != D:
        raise ValueError("bending_energy() 'u' must be tensor of shape (N, D, ..., X)")
    kwargs = dict(mode=mode or "sobel", sigma=sigma, spacing=spacing, stride=stride)
    which = FlowDerivativeKeys.all(spatial_dims=D, order=2)
    which = sorted(FlowDerivativeKeys.unique(which))
    deriv = flow_derivatives(u, which=which, **kwargs)
    loss: Optional[Tensor] = None
    for key, value in deriv.items():
        value = value.square_()
        if FlowDerivativeKeys.is_mixed(key):
            value = value.mul_(2)
        loss = value if loss is None else loss.add_(value)
    assert loss is not None
    loss = reduce_loss(loss, reduction)
    return loss


be_loss = bending_loss
bending_energy = bending_loss


def bspline_bending_loss(
    data: Tensor, stride: ScalarOrTuple[int] = 1, reduction: str = "mean"
) -> Tensor:
    r"""Evaluate bending energy of cubic B-spline function, e.g., spatial free-form deformation.

    Deprecated, use :func:`bending_loss()` with ``mode='bspline'`` instead.

    Args:
        data: Cubic B-spline interpolation coefficients as tensor of shape ``(N, C, ..., X)``.
        stride: Number of points between control points at which to evaluate bending energy, plus one.
            If a sequence of values is given, these must be the strides for the different spatial
            dimensions in the order ``(sx, ...)``. A stride of 1 is equivalent to evaluating bending
            energy only at the usually coarser resolution of the control point grid. It should be noted
            that the stride need not match the stride used to densely sample the spline deformation field
            at a given fixed target image resolution.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Bending energy of cubic B-spline.

    """
    return bending_loss(data, mode="bspline", stride=stride, reduction=reduction)


bspline_be_loss = bspline_bending_loss
bspline_bending_energy = bspline_bending_loss


def curvature_loss(
    u: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Loss term based on unmixed 2nd order spatial derivatives of vector fields.

    References:
        Fischer & Modersitzki (2003). Curvature based image registration.
            Journal Mathematical Imaging and Vision, 18(1), 81–85.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        mode: Method used to approximate :func:`flow_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Curvature loss of vector fields.

    """
    if u.ndim < 4:
        # No loss for homogeneous coordinate transformations
        if reduction == "none":
            raise NotImplementedError(
                "curvature_loss() not implemented for linear transformation and reduction='none'"
            )
        return torch.tensor(0, dtype=u.dtype, device=u.device)
    N = u.shape[0]
    D = u.shape[1]
    if u.ndim - 2 != D:
        raise ValueError(f"curvature_loss() 'u' must be tensor of shape (N, {u.ndim - 2}, ..., X)")
    kwargs = dict(mode=mode or "sobel", sigma=sigma, spacing=spacing, stride=stride)
    which = FlowDerivativeKeys.curvature(spatial_dims=D)
    deriv = flow_derivatives(u, which=which, **kwargs)
    loss = torch.zeros((N, D) + u.shape[2:], dtype=u.dtype, device=u.device)
    for i, j in itertools.product(range(D), repeat=2):
        loss.narrow(1, i, 1).add_(deriv[FlowDerivativeKeys.symbol(i, j, j)])
    loss = loss.square_().sum(dim=1, keepdim=True)
    loss = reduce_loss(loss, reduction)
    return loss.mul_(0.5)


def diffusion_loss(
    u: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Diffusion regularization loss.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        mode: Method used to approximate :func:`flow_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Diffusion regularization loss of vector fields.

    """
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    loss = grad_loss(u, p=2, q=1, reduction=reduction, **kwargs)
    return loss.mul_(0.5)


def divergence_loss(
    u: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Loss term encouraging divergence-free vector fields.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        mode: Method used to approximate :func:`flow_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Divergence loss of vector fields.

    """
    if u.ndim < 4:
        # No loss for homogeneous coordinate transformations
        if reduction == "none":
            raise NotImplementedError(
                "divergence_loss() not implemented for linear transformation and reduction='none'"
            )
        return torch.tensor(0, dtype=u.dtype, device=u.device)
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    loss = divergence(u, **kwargs).square_()
    loss = reduce_loss(loss, reduction)
    return loss.mul_(0.5)


def lame_parameters(
    material_name: Optional[str] = None,
    first_parameter: Optional[float] = None,
    second_parameter: Optional[float] = None,
    shear_modulus: Optional[float] = None,
    poissons_ratio: Optional[float] = None,
    youngs_modulus: Optional[float] = None,
) -> Tuple[float, float]:
    r"""Get Lame parameters of linear elasticity given different quantities.

    Args:
        material_name: Name of material preset. Cannot be used in conjunction with other arguments.
        first_parameter: Lame's first parameter.
        second_parameter: Lame's second parameter, i.e., shear modulus.
        shear_modulus: Shear modulus, i.e., Lame's second parameter.
        poissons_ratio: Poisson's ratio.
        youngs_modulus: Young's modulus.

    Returns:
        lambda: Lame's first parameter.
        mu: Lame's second parameter.

    """
    RUBBER_POISSONS_RATIO = 0.4999
    RUBBER_SHEAR_MODULUS = 0.0006
    # Derive unspecified Lame parameters from any combination of two given quantities
    # (cf. conversion table at https://en.wikipedia.org/wiki/Young%27s_modulus#External_links)
    kwargs = {
        name: value
        for name, value in zip(
            [
                "first_parameter",
                "second_parameter",
                "shear_modulus",
                "poissons_ratio",
                "youngs_modulus",
            ],
            [first_parameter, second_parameter, poissons_ratio, youngs_modulus, shear_modulus],
        )
        if value is not None
    }
    # Default values for different materials (cf. Wikipedia pages for Poisson's ratio and shear modulus)
    if material_name:
        if kwargs:
            raise ValueError(
                "lame_parameters() 'material_name' cannot be specified in combination with other quantities"
            )
        if material_name == "rubber":
            poissons_ratio = RUBBER_POISSONS_RATIO
            shear_modulus = RUBBER_SHEAR_MODULUS
        else:
            raise ValueError(f"lame_parameters() unknown 'material_name': {material_name}")
    elif len(kwargs) != 2:
        raise ValueError(
            "lame_parameters() specify 'material_name' or exactly two parameters, got: "
            + ", ".join(f"{k}={v}" for k, v in kwargs.items())
        )
    if second_parameter is None:
        second_parameter = shear_modulus
    elif shear_modulus is None:
        shear_modulus = second_parameter
    else:
        raise ValueError(
            "lame_parameters() 'second_parameter' and 'shear_modulus' are mutually exclusive"
        )
    if first_parameter is None:
        if shear_modulus is None:
            if poissons_ratio is not None and youngs_modulus is not None:
                first_parameter = (
                    poissons_ratio * youngs_modulus / ((1 + poissons_ratio)(1 - 2 * poissons_ratio))
                )
                second_parameter = youngs_modulus / (2 * (1 + poissons_ratio))
        elif youngs_modulus is None:
            if poissons_ratio is None:
                poissons_ratio = RUBBER_POISSONS_RATIO
            first_parameter = 2 * shear_modulus * poissons_ratio / (1 - 2 * poissons_ratio)
        else:
            first_parameter = (
                shear_modulus
                * (youngs_modulus - 2 * shear_modulus)
                / (3 * shear_modulus - youngs_modulus)
            )
    elif second_parameter is None:
        if youngs_modulus is None:
            if poissons_ratio is None:
                poissons_ratio = RUBBER_POISSONS_RATIO
            second_parameter = first_parameter * (1 - 2 * poissons_ratio) / (2 * poissons_ratio)
        else:
            r = math.sqrt(
                youngs_modulus**2
                + 9 * first_parameter**2
                + 2 * youngs_modulus * first_parameter
            )
            second_parameter = youngs_modulus - 3 * first_parameter + r / 4
    if first_parameter is None or second_parameter is None:
        raise NotImplementedError(
            "lame_parameters() deriving Lame parameters from: "
            + ", ".join(f"'{name}'" for name in kwargs.keys())
        )
    if first_parameter < 0:
        raise ValueError("lame_parameter() 'first_parameter' is negative")
    elif first_parameter < 1e-9:
        first_parameter = 0
    if second_parameter < 0:
        raise ValueError("lame_parameter() 'second_parameter' is negative")
    elif second_parameter < 1e-9:
        second_parameter = 0
    return first_parameter, second_parameter


def elasticity_loss(
    u: Tensor,
    material_name: Optional[str] = None,
    first_parameter: Optional[float] = None,
    second_parameter: Optional[float] = None,
    shear_modulus: Optional[float] = None,
    poissons_ratio: Optional[float] = None,
    youngs_modulus: Optional[float] = None,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Loss term based on Navier-Cauchy PDE of linear elasticity.

    References:
        Fischer & Modersitzki, 2004, A unified approach to fast image registration and a new
            curvature based registration technique.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        material_name: Name of material preset. Cannot be used in conjunction with other arguments.
        first_parameter: Lame's first parameter.
        second_parameter: Lame's second parameter, i.e., shear modulus.
        shear_modulus: Shear modulus, i.e., Lame's second parameter.
        poissons_ratio: Poisson's ratio.
        youngs_modulus: Young's modulus.
        mode: Method used to approximate :func:`flow_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Linear elasticity loss of vector field.

    """
    lambd, mu = lame_parameters(
        material_name=material_name,
        first_parameter=first_parameter,
        second_parameter=second_parameter,
        shear_modulus=shear_modulus,
        poissons_ratio=poissons_ratio,
        youngs_modulus=youngs_modulus,
    )
    if u.ndim < 4:
        # No loss for homogeneous coordinate transformations
        if reduction == "none":
            raise NotImplementedError(
                "elasticity_loss() not implemented for linear transformation and reduction='none'"
            )
        return torch.tensor(0, dtype=u.dtype, device=u.device)
    N = u.shape[0]
    D = u.shape[1]
    if u.ndim - 2 != D:
        raise ValueError(f"elasticity_loss() 'u' must be tensor of shape (N, {u.ndim - 2}, ..., X)")
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    which = FlowDerivativeKeys.jacobian(spatial_dims=D)
    deriv = flow_derivatives(u, which=which, **kwargs)
    loss = torch.zeros((N, 1) + u.shape[2:], dtype=u.dtype, device=u.device)
    if lambd != 0:
        for i in range(D):
            loss = loss.add_(deriv[FlowDerivativeKeys.symbol(i, i)])
        loss = loss.square_().mul_(lambd / 2)
    if mu != 0:
        for j, k in itertools.product(range(D), repeat=2):
            djk = deriv[FlowDerivativeKeys.symbol(j, k)]
            dkj = deriv[FlowDerivativeKeys.symbol(k, j)]
            loss = loss.add_(djk.add(dkj).square_().mul_(mu / 4))
    loss = reduce_loss(loss, reduction)
    return loss


def total_variation_loss(
    u: Tensor,
    mode: Optional[str] = None,
    sigma: Optional[float] = None,
    spacing: Optional[Union[Scalar, Array]] = None,
    stride: Optional[ScalarOrTuple] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Total variation regularization loss.

    Args:
        u: Batch of vector fields as tensor of shape ``(N, D, ..., X)``. When a tensor with less than
            four dimensions is given, it is assumed to be a linear transformation and zero is returned.
        mode: Method used to approximate :func:`flow_derivatives()`.
        sigma: Standard deviation of Gaussian in grid units used to smooth vector field.
        spacing: Step size to use when computing finite differences.
        stride: Number of output grid points between control points plus one for ``mode='bspline'``.
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Total variation regularization loss of vector fields.

    """
    kwargs = dict(mode=mode, sigma=sigma, spacing=spacing, stride=stride)
    return grad_loss(u, p=1, q=1, reduction=reduction, **kwargs)


tv_loss = total_variation_loss


def inverse_consistency_loss(
    forward: Tensor,
    inverse: Tensor,
    grid: Optional[Grid] = None,
    margin: Union[int, float] = 0,
    mask: Optional[Tensor] = None,
    units: str = "cube",
    reduction: str = "mean",
) -> Tensor:
    r"""Evaluate inverse consistency error.

    This function expects forward and inverse coordinate maps to be with respect to the unit cube
    of side length 2 as defined by the domain and codomain ``grid`` (see also ``Grid.axes()``).

    Args:
        forward: Tensor representation of spatial transformation.
        inverse: Tensor representation of inverse transformation.
        grid: Coordinate domain and codomain of forward transformation.
        margin: Number of ``grid`` points to ignore when computing mean error. If type of the
            argument is ``int``, this number of points are dropped at each boundary in each dimension.
            If a ``float`` value is given, it must be in [0, 1) and denote the percentage of sampling
            points to drop at each border. Inverse consistency of points near the domain boundary is
            affected by extrapolation and excluding these may be preferrable. See also ``mask``.
        mask: Foreground mask as tensor of shape ``(N, 1, ..., X)`` with size matching ``forward``.
            Inverse consistency errors at target grid points with a zero mask value are ignored.
        units: Compute mean inverse consistency error in specified units: ``cube`` with respect to
            normalized grid cube coordinates, ``voxel`` in voxel units, or in ``world`` units (mm).
        reduction: Specifies the reduction to apply to the output: 'none' | 'mean' | 'sum'.

    Returns:
        Inverse consistency error.

    """
    if not isinstance(forward, Tensor):
        raise TypeError("inverse_consistency_loss() 'forward' must be tensor")
    if not isinstance(inverse, Tensor):
        raise TypeError("inverse_consistency_loss() 'inverse' must be tensor")
    if not isinstance(margin, (int, float)):
        raise TypeError("inverse_consistency_loss() 'margin' must be int or float")
    if grid is None:
        if forward.ndim < 4:
            if inverse.ndim < 4:
                raise ValueError(
                    "inverse_consistency_loss() 'grid' required when both transforms are affine"
                )
            grid = Grid(shape=inverse.shape[2:])
        else:
            grid = Grid(shape=forward.shape[2:])
    # Compute inverse consistency error for each grid point
    x = grid.coords(dtype=forward.dtype, device=forward.device).unsqueeze(0)
    y = transform_grid(forward, x, align_corners=grid.align_corners())
    y = transform_points(inverse, y, align_corners=grid.align_corners())
    error = y - x
    # Set error outside foreground mask to zero
    if mask is not None:
        if not isinstance(mask, Tensor):
            raise TypeError("inverse_consistency_loss() 'mask' must be tensor")
        if mask.ndim != grid.ndim + 2:
            raise ValueError(
                f"inverse_consistency_loss() 'mask' must be {grid.ndim + 2}-dimensional"
            )
        if mask.shape[1] != 1:
            raise ValueError("inverse_consistency_loss() 'mask' must have shape (N, 1, ..., X)")
        if mask.shape[0] != 1 and mask.shape[0] != error.shape[0]:
            raise ValueError(
                f"inverse_consistency_loss() 'mask' batch size must be 1 or {error.shape[0]}"
            )
        error[move_dim(mask == 0, 1, -1).expand_as(error)] = 0
    # Discard error at grid boundary
    if margin > 0:
        if isinstance(margin, float):
            if margin < 0 or margin >= 1:
                raise ValueError(
                    f"inverse_consistency_loss() 'margin' must be in [0, 1), got {margin}"
                )
            m = [int(margin * n) for n in grid.size()]
        else:
            m = [max(0, int(margin))] * grid.ndim
        subgrid = tuple(reversed([slice(i, n - i) for i, n in zip(m, grid.size())]))
        error = error[(slice(0, error.shape[0]),) + subgrid + (slice(0, grid.ndim),)]
    # Scale differences by respective error units
    if units in ("voxel", "world"):
        error = denormalize_flow(error, size=grid.size(), channels_last=True)
        if units == "world":
            error *= grid.spacing().to(error)
    # Calculate error norm
    error: Tensor = error.norm(p=2, dim=-1)
    # Reduce error if requested
    if reduction != "none":
        count = error.numel()
        error = error.sum()
        if reduction == "mean" and mask is not None:
            count = (mask != 0).sum()
        error /= count
    return error


def elementwise_loss(
    name: str,
    loss_fn: ElementwiseLoss,
    input: Tensor,
    target: Tensor,
    mask: Optional[Tensor] = None,
    norm: Optional[Union[float, Tensor]] = None,
    reduction: str = "mean",
) -> Tensor:
    r"""Evaluate, aggregate, and normalize elementwise loss, optionally within masked region.

    Args:
        input: Source image sampled on ``target`` grid.
        target: Target image with same shape as ``input``.
        mask: Multiplicative mask with same shape as ``input``.
        norm: Positive factor by which to divide loss value.
        reduction: Whether to compute "mean" or "sum" over all grid points.
            If "none", output tensor shape is equal to the shape of the input tensors.

    Returns:
        Aggregated normalized loss value.

    """
    if not isinstance(input, Tensor):
        raise TypeError(f"{name}() 'input' must be tensor")
    if not isinstance(target, Tensor):
        raise TypeError(f"{name}() 'target' must be tensor")
    if input.shape != target.shape:
        raise ValueError(f"{name}() 'input' must have same shape as 'target'")
    if mask is None:
        loss = loss_fn(input, target, reduction=reduction)
    else:
        loss = loss_fn(input, target, reduction="none")
        loss = masked_loss(loss, mask, name)
        loss = reduce_loss(loss, reduction, mask)
    if norm is not None:
        norm = torch.as_tensor(norm, dtype=loss.dtype, device=loss.device).squeeze()
        if not norm.ndim == 0:
            raise ValueError(f"{name}() 'norm' must be scalar")
        if norm > 0:
            loss = loss.div_(norm)
    return loss


def masked_loss(
    loss: Tensor,
    mask: Optional[Tensor] = None,
    name: Optional[str] = None,
    inplace: bool = False,
) -> Tensor:
    r"""Multiply loss with an optionally specified spatial mask."""
    if mask is None:
        return loss
    if not name:
        name = "masked_loss"
    if not isinstance(mask, Tensor):
        raise TypeError(f"{name}() 'mask' must be tensor")
    if mask.shape[0] != 1 and mask.shape[0] != loss.shape[0]:
        raise ValueError(f"{name}() 'mask' must have same batch size as 'target' or batch size 1")
    if mask.shape[1] != 1 and mask.shape[1] != loss.shape[1]:
        raise ValueError(f"{name}() 'mask' must have same number of channels as 'target' or only 1")
    if mask.shape[2:] != loss.shape[2:]:
        raise ValueError(f"{name}() 'mask' must have same spatial shape as 'target'")
    if inplace:
        loss = loss.mul_(mask)
    else:
        loss = loss.mul(mask)
    return loss


def reduce_loss(loss: Tensor, reduction: str = "mean", mask: Optional[Tensor] = None) -> Tensor:
    r"""Reduce loss computed at each grid point."""
    if reduction not in ("mean", "sum", "none"):
        raise ValueError("reduce_loss() 'reduction' must be 'mean', 'sum' or 'none'")
    if reduction == "none":
        return loss
    if mask is None:
        return loss.mean() if reduction == "mean" else loss.sum()
    value = loss.sum()
    if reduction == "mean":
        numel = mask.expand_as(loss).sum()
        value = value.div_(numel)
    return value



r"""Client for AWS Simple Storage Service (S3)."""

from __future__ import annotations

from contextlib import contextmanager
from enum import Enum
import io
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Set, Tuple, Union

import boto3
from botocore.exceptions import ClientError

from .config import S3Config

S3ClientError = ClientError
PathStr = Union[Path, str]


def match_all(key: str) -> bool:
    r"""Default 'match' function used by 'S3Client.download_files()'."""
    return True


class S3Client(object):
    r"""Client for AWS Simple Storage Service (S3)."""

    class Operation(Enum):
        r"""Enumeration of permissible S3 operations.

        Note that these permissions restrict the set of allowed operations and
        is independent of S3 object permissions which are configured in AWS S3.

        """

        READ = "r"
        WRITE = "w"
        DELETE = "d"

    _default: List[S3Client] = []

    @classmethod
    def default(cls, client: Optional[S3Client] = None) -> S3Client:
        r"""Default client instance.

        This function is not thread-safe if not called first by the main thread.

        Args:
            client: If ``None``, a new client instance is created if no current default client exists.
                Otherwise, the given client instance replaces the current default instance.

        Returns:
            Current default client instance. If no default client instance has been set before,
                a new instance with default configuration is created.

        """
        if client is not None:
            if cls._default:
                cls._default[-1] = client
            else:
                cls._default.append(client)
        elif not cls._default:
            cls._default.append(cls())
        return cls._default[-1]

    @classmethod
    @contextmanager
    def init_default(cls, *args, **kwargs) -> Generator[S3Client, None, None]:
        r"""Set new default client to use in current context.

        This function is not thread-safe. Only the main thread may set the default client
        instance to use by worker threads. Any multi-threaded operations should share
        a client instance, a reference to which should be provided by the main thread.
        Nested ``with`` statements using this context manager function are supported,
        as long as these occur in the main thread only.

        Args:
            *args: Positional arguments for client ``__init__`` function call.
            **kwargs: Keyword arguments for client ``__init__`` function call.

        Returns:
            client (S3Client): Generator function that returns the connected default client
                instance and resets the default to the previously active one upon completion.

        """
        client = cls(*args, **kwargs)
        client.connect()
        cls._default.append(client)
        try:
            yield client
        finally:
            client = cls._default.pop()
            client.close()

    def __init__(self, config: Optional[S3Config] = None, **kwargs) -> None:
        r"""Initialize S3 client.

        Args:
            config: Client configuration.
            **kwargs: Individual client configuration settings.

        """
        if config is None:
            config = S3Config()
        self._config = config._replace(**kwargs)
        self._client = None
        self._depth = 0

    @property
    def exceptions(self):
        r"""Get boto3 exceptions for connected client."""
        if self._client is None:
            raise AssertionError(
                f"{type(self).__name__} must be connected to access exceptions."
                " Mabye use botocore.exceptions module instead."
            )
        return self._client.exceptions

    @classmethod
    def from_arg(cls, arg: Union[S3Config, S3Client, Dict[str, Any], None]) -> S3Client:
        r"""Get client instance given function argument.

        Args:
            arg: Function argument.

        Returns:
            client: S3 client instance. If ``arg`` is an S3Client instance,
                this instance is returned. If ``arg`` is of type S3Config or dict, a new
                instance is returned, which has been initialized with this configuration.
                If ``arg`` is ``None``, the default client instance is returned.

        """
        if isinstance(arg, cls):
            return arg
        if isinstance(arg, S3Config):
            return cls(arg)
        if isinstance(arg, dict):
            return cls(**arg)
        return cls.default()

    @property
    def config(self) -> S3Config:
        r"""Get client configuration object."""
        return self._config

    @property
    def ops(self) -> Set[S3Client.Operation]:
        r"""Get list of permissible client operations."""
        return set(S3Client.Operation(c) for c in self.config.mode)

    def connect(self) -> S3Client:
        r"""Establish connection with AWS S3."""
        if self._client is None:
            self._client = boto3.client(
                "s3", region_name=self.config.region, verify=self.config.verify
            )
        return self

    def close(self) -> None:
        r"""Close connection with AWS S3."""
        self._client = None

    def is_closed(self) -> bool:
        r"""Check if S3 connection is closed."""
        return self._client is None

    def is_open(self) -> bool:
        r"""Check if client connection is open."""
        return not self.is_closed()

    def __enter__(self) -> S3Client:
        r"""Ensure client connection is open when entering context.

        This function increments a context depth counter in a non-thread-safe way.
        Only the main thread should create client connection and pass these on to
        worker threads that require access to this resource.

        """
        self.connect()
        self._depth = max(1, self._depth + 1)
        return self

    def __exit__(self, *exc) -> None:
        r"""Close connection when exiting outermost context."""
        self._depth = max(0, self._depth - 1)
        if self._depth == 0:
            self.close()

    def exists(self, bucket: str, key: str) -> bool:
        r"""Check if specified S3 object exists.

        Args:
            bucket: Bucket name.
            key: S3 object key. To query the existence of a folder, the
                key must end with a forward slash character ('/').

        Returns:
            Whether S3 object exists in specified bucket.

        """
        assert not self.is_closed(), "client connection required"
        if key.endswith("/"):
            resp = self._client.list_objects_v2(Bucket=bucket, Prefix=key)
            flag = bool(resp.get("Contents", []))
        else:
            flag = False
            try:
                self._client.head_object(Bucket=bucket, Key=key)
                flag = True
            except self._client.exceptions.ClientError as error:
                error_info = getattr(error, "response", {}).get("Error", {})
                if error_info.get("Code") != "404":
                    raise error
            except self._client.exceptions.NoSuchKey:
                pass
        return flag

    def keys(self, bucket: str, prefix: Optional[str] = None) -> Generator[str, None, None]:
        r"""List S3 objects with specified key prefix.

        Args:
            bucket: Bucket name.
            prefix: Common key prefix.

        Returns:
            Generator of S3 object keys.

        Raises:
            PermissionError: If read operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        if S3Client.Operation.READ not in self.ops:
            raise PermissionError("S3 client has no read permissions")
        kwargs = {"Bucket": bucket}
        if prefix:
            kwargs["Prefix"] = prefix
        while True:
            resp: dict = self._client.list_objects_v2(**kwargs)
            for obj in resp.get("Contents", []):
                yield obj["Key"]
            try:
                kwargs["ContinuationToken"] = resp["NextContinuationToken"]
            except KeyError:
                break

    def iterdir(self, bucket: str, prefix: Optional[str] = None) -> Generator[str, None, None]:
        r"""List S3 objects with specified key prefix, excluding subfolder contents.

        Args:
            bucket: Bucket name.
            prefix: Common key prefix. To list the contents of a folder, the
                key prefix must end with a forward slash character ('/'), otherwise
                the result will only contain the folder itself.

        Returns:
            Generator of S3 object keys, where subfolders are represented
                by keys ending with a forward slash character ('/').

        Raises:
            PermissionError: If read operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        if S3Client.Operation.READ not in self.ops:
            raise PermissionError("S3 client has no read permissions")
        kwargs = {"Bucket": bucket, "Delimiter": "/"}
        if prefix:
            kwargs["Prefix"] = prefix
        while True:
            resp: dict = self._client.list_objects_v2(**kwargs)
            for obj in resp.get("Contents", []):
                yield obj["Key"]
            for obj in resp.get("CommonPrefixes", []):
                yield obj["Prefix"]
            try:
                kwargs["ContinuationToken"] = resp["NextContinuationToken"]
            except KeyError:
                break

    def listdir(self, bucket: str, prefix: Optional[str] = None) -> Generator[str, None, None]:
        r"""List names of S3 objects whose keys match a given prefix, excluding subfolder contents.

        This convenience function uses ``iterdir`` to obtain the keys of the S3 objects matching
        a given ``prefix`` that correspond to the respective folder, and removes the folder key
        prefix from the resulting object keys. It thus behaves similar to ``os.listdir``.

        Args:
            bucket: Bucket name.
            prefix: Common key prefix. To list the contents of a folder, the
                key prefix must end with a forward slash character ('/'), otherwise
                the result will only contain the folder itself.

        Returns:
            Generator of S3 object keys, where subfolders are represented
                by keys ending with a forward slash character ('/').

        Raises:
            PermissionError: If read operations have not been enabled for this client.

        """
        skip = (prefix or "").rfind("/") + 1
        for key in self.iterdir(bucket, prefix):
            yield key[skip:]

    def read_bytes(self, bucket: str, key: str) -> bytes:
        r"""Download binary object data.

        Args:
            bucket: Bucket name.
            key: S3 object key.

        Returns:
            data: S3 object data.

        Raises:
            PermissionError: If read operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        assert key, "S3 object key must be specified"
        if S3Client.Operation.READ not in self.ops:
            raise PermissionError("S3 client has no read permissions")
        buffer = io.BytesIO()
        config = self._config.transfer_config()
        self._client.download_fileobj(Bucket=bucket, Key=key, Fileobj=buffer, Config=config)
        return buffer.getvalue()

    def write_bytes(self, bucket: str, key: str, data: bytes) -> None:
        r"""Upload binary object data.

        Args:
            bucket: Bucket name.
            key: S3 object key.
            data: S3 object data.

        Raises:
            PermissionError: If write operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        assert key, "S3 object key must be specified"
        if S3Client.Operation.WRITE not in self.ops:
            raise PermissionError("S3 client has no write permissions")
        buffer = io.BytesIO(data)
        config = self._config.transfer_config()
        self._client.upload_fileobj(Fileobj=buffer, Bucket=bucket, Key=key, Config=config)

    def read_text(self, bucket: str, key: str, encoding: Optional[str] = None) -> str:
        r"""Download text file content.

        Args:
            bucket: Bucket name.
            key: S3 object key.
            encoding: Text encoding.

        Returns:
            text: Decoded text.

        Raises:
            PermissionError: If read operations have not been enabled for this client.

        """
        data = self.read_bytes(bucket, key)
        return data.decode(encoding) if encoding else data.decode()

    def write_text(self, bucket: str, key: str, text: str, encoding: Optional[str] = None) -> None:
        r"""Upload text file content.

        Args:
            bucket: Bucket name.
            key: S3 object key.
            text: S3 object data.
            encoding: Text encoding.

        Raises:
            PermissionError: If write operations have not been enabled for this client.

        """
        data = text.encode(encoding) if encoding else text.encode()
        self.write_bytes(bucket=bucket, key=key, data=data)

    def download_file(self, bucket: str, key: str, path: PathStr, overwrite: bool = True) -> None:
        r"""Download S3 object to local file.

        Args:
            bucket: Bucket name.
            key: S3 object key.
            path: Local file path or path of existing output directory.
            overwrite: Whether to overwrite existing local file.

        Raises:
            PermissionError: If read operations have not been enabled for this client.
            FileExistsError: If local ``path`` exists and ``overwrite=False``.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        assert key, "S3 object key must be specified"
        if S3Client.Operation.READ not in self.ops:
            raise PermissionError("S3 client has no read permissions")
        path = Path(path).absolute()
        if path.is_dir():
            path = path.joinpath(key.rsplit("/", 1)[-1])
        if not overwrite and path.is_file():
            raise FileExistsError(
                "Use overwrite=True to force overwriting existing local file '{}'".format(path)
            )
        data = self.read_bytes(bucket, key)
        try:
            path.unlink()
        except FileNotFoundError:
            path.parent.mkdir(parents=True, exist_ok=True)
        try:
            path.write_bytes(data)
        except (FileNotFoundError, PermissionError):
            raise
        except Exception as e:
            try:
                path.unlink()
            except Exception:
                pass
            raise e

    def download_files(
        self,
        bucket: str,
        prefix: str,
        path: PathStr,
        match: Optional[Callable[[str], bool]] = None,
        overwrite: bool = True,
    ) -> Tuple[int, int]:
        r"""Download S3 objects to local directory.

        Args:
            bucket: Bucket name.
            prefix: Common key prefix.
            path: Local directory path. If the directory exists, downloaded
                files are merged into this existing directory and subdirectories.
            match: Filter function that takes a subkey without ``prefix`` and
                returns either True or False. If False, the corresponding file is skipped.
            overwrite: Whether to overwrite existing local files.

        Returns:
            total: Number of matching S3 objects.
            count: Number of actually downloaded files.

        Raises:
            PermissionError: If read operations have not been enabled for this client.

        """
        total = 0
        count = 0
        skip = prefix.rfind("/") + 1
        if match is None:
            match = match_all
        for key in self.keys(bucket=bucket, prefix=prefix):
            subkey = key[skip:]
            if subkey and match(subkey):
                total += 1
                try:
                    self.download_file(
                        bucket=bucket,
                        key=key,
                        path=path.joinpath(subkey),
                        overwrite=overwrite,
                    )
                    count += 1
                except FileExistsError:
                    pass
        return total, count

    def upload_file(self, path: PathStr, bucket: str, key: str, overwrite: bool = True) -> None:
        r"""Upload local file to S3.

        Args:
            path: Local file path.
            bucket: Bucket name.
            key: S3 object key.
            overwrite: Whether to overwrite existing S3 object.

        Raises:
            PermissionError: If write operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        assert key, "S3 object key must be specified"
        assert not key.endswith("/"), "S3 object key must not end with forward slash"
        if S3Client.Operation.WRITE not in self.ops:
            raise PermissionError("S3 client has no write permissions")
        if not overwrite and self.exists(bucket=bucket, key=key):
            raise FileExistsError(
                "Use overwrite=True to force overwriting "
                + "existing S3 object '{k}' in bucket {b}".format(k=key, b=bucket)
            )
        path = Path(path).absolute()
        self.write_bytes(bucket=bucket, key=key, data=path.read_bytes())

    def upload_files(
        self, path: PathStr, bucket: str, prefix: Optional[str] = None, overwrite: bool = True
    ) -> Tuple[int, int]:
        r"""Upload local directory to S3.

        Args:
            path: Local directory path.
            bucket: Bucket name.
            prefix: Common S3 object key prefix. If ``None`` or empty string,
                the content of the local directory is uploaded to the specified
                ``bucket`` without any object key prefix, i.e., the files will
                be located in S3 directly underneath the destination bucket.
            overwrite: Whether to overwrite existing S3 objects.

        Returns:
            total: Number of local files.
            count: Number of uploaded files.

        Raises:
            PermissionError: If write operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        total = 0
        count = 0
        if prefix is None:
            prefix = "/"
        if not prefix.endswith("/"):
            prefix += "/"
        path = Path(path).absolute()
        for child in path.iterdir():
            key = prefix + child.name
            if child.is_dir():
                tot, cnt = self.upload_files(
                    path=child, bucket=bucket, prefix=key, overwrite=overwrite
                )
                total += tot
                count += cnt
            else:
                try:
                    self.upload_file(path=child, bucket=bucket, key=key, overwrite=overwrite)
                    count += 1
                except FileExistsError:
                    pass
                total += 1
        return total, count

    def delete_file(self, bucket: str, key: str) -> int:
        r"""Delete S3 object.

        Args:
            bucket: Bucket name.
            key: S3 object key.

        Returns:
            count: Number of deleted S3 objects (0 or 1).

        Raises:
            PermissionError: If delete operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        if S3Client.Operation.DELETE not in self.ops:
            raise PermissionError("S3 client has no permission to delete objects")
        resp = self._client.delete_object(Bucket=bucket, Key=key)
        return 1 if "DeleteMarker" in resp else 0

    def delete_files(self, bucket: str, prefix: str) -> int:
        r"""Delete S3 objects.

        Args:
            bucket: Bucket name.
            prefix: Common key prefix.

        Returns
            count: Number of deleted S3 objects.

        Raises:
            PermissionError: If delete operations have not been enabled for this client.

        """
        assert not self.is_closed(), "client connection required"
        assert bucket, "S3 bucket must be specified"
        if S3Client.Operation.DELETE not in self.ops:
            raise PermissionError("S3 client has no permission to delete objects")
        count = 0
        list_kwargs = {"Bucket": bucket, "Prefix": prefix}
        del_kwargs = {"Bucket": bucket, "Delete": {"Objects": []}}
        while True:
            list_resp = self._client.list_objects_v2(**list_kwargs)
            del_kwargs["Delete"]["Objects"] = [
                {"Key": obj["Key"]} for obj in list_resp.get("Contents", [])
            ]
            if del_kwargs["Delete"]["Objects"]:
                count += len(self._client.delete_objects(**del_kwargs)["Deleted"])
            try:
                list_resp["ContinuationToken"] = list_resp["NextContinuationToken"]
            except KeyError:
                break
        return count



r"""Sequential network layers with skip connections."""

from typing import Any, Callable, Mapping, Union, overload

from torch import Tensor
from torch.nn import Identity, Module, ModuleDict, Sequential

from deepali.modules.mixins import ReprWithCrossReferences

from ..layers.join import JoinFunc, join_func


# fmt: off
__all__ = (
    "DenseBlock",
    "Shortcut",
    "SkipConnection",
    "SkipFunc",
)
# fmt: on


SkipFunc = Callable[[Tensor], Tensor]


class DenseBlock(ReprWithCrossReferences, Module):
    r"""Subnetwork with dense skip connections."""

    @overload
    def __init__(self, *args: Module, join: Union[str, JoinFunc], dim: int) -> None:
        ...

    @overload
    def __init__(self, arg: Mapping[str, Module], join: Union[str, JoinFunc], dim: int) -> None:
        ...

    def __init__(self, *args: Any, join: Union[str, JoinFunc] = "cat", dim: int = 1) -> None:
        super().__init__()
        if len(args) == 1 and isinstance(args[0], Mapping):
            layers = args[0]
        else:
            layers = {str(i): m for i, m in enumerate(args)}
        self.layers = ModuleDict(layers)
        self.join = join_func(join, dim=dim)
        self.is_associative = join in ("add", "cat", "concat")

    def forward(self, x: Tensor) -> Tensor:
        y, ys = x, [x]
        join = self.join
        is_associative = self.is_associative
        for module in self.layers.values():
            x = join(ys)
            y = module(x)
            ys = [x, y] if is_associative else [*ys, y]
        return y


class SkipConnection(ReprWithCrossReferences, Module):
    r"""Combine input with subnetwork output along a single skip connection."""

    @overload
    def __init__(
        self,
        *args: Module,
        name: str = "func",
        skip: Union[str, SkipFunc] = "identity",
        join: Union[str, JoinFunc] = "cat",
        dim: int = 1,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg: Mapping[str, Module],
        name: str = "func",
        skip: Union[str, SkipFunc] = "identity",
        join: Union[str, JoinFunc] = "cat",
        dim: int = 1,
    ) -> None:
        ...

    def __init__(
        self,
        *args: Any,
        name: str = "func",
        skip: Union[str, SkipFunc] = "identity",
        join: Union[str, JoinFunc] = "cat",
        dim: int = 1,
    ) -> None:
        super().__init__()
        if len(args) == 1 and isinstance(args[0], Module):
            func = args[0]
        else:
            func = Sequential(*args)
        self.name = name
        if skip in (None, "identity"):
            skip = Identity()
        elif not callable(skip):
            raise ValueError("SkipConnection() 'skip' must be 'identity', callable, or None")
        self.skip = skip
        self.join = join_func(join, dim=dim)
        self._modules[self.name] = func

    @property
    def func(self) -> Module:
        return self._modules[self.name]

    @property
    def shortcut(self) -> SkipFunc:
        return self.skip

    def forward(self, x: Tensor) -> Tensor:
        a = self.skip(x)
        b = self.func(x)
        if not isinstance(a, Tensor):
            raise TypeError("SkipConnection() 'skip' function must return a Tensor")
        if not isinstance(b, Tensor):
            raise TypeError("SkipConnection() module must return a Tensor")
        c = self.join([b, a])
        if not isinstance(c, Tensor):
            raise TypeError("SkipConnection() 'join' function must return a Tensor")
        return c


Shortcut = SkipConnection



r"""Functions to resample image data."""

from typing import Optional, Union

import numpy as np
import scipy.interpolate
import scipy.ndimage
import scipy.spatial
import SimpleITK as sitk

from .grid import GridAttrs, image_grid_attributes


def resample_image(
    image: sitk.Image,
    reference: Union[GridAttrs, sitk.Image],
    interpolator: int = sitk.sitkLinear,
    padding_value: float = 0,
) -> sitk.Image:
    r"""Interpolate image values at grid points of reference image.

    Args:
        image: Scalar or vector-valued image to evaluate at the specified points.
        reference: Sampling grid on which to evaluate interpolated image. If a path is specified,
        interpolator: Enumeration value of SimpleITK interpolator to use.
        padding_value: Output value when sampling point is outside the input image domain.

    Returns:
        output: Image interpolated at reference grid points.

    """
    resampler = sitk.ResampleImageFilter()
    resampler.SetInterpolator(interpolator)
    resampler.SetDefaultPixelValue(padding_value)
    if isinstance(reference, sitk.Image):
        resampler.SetReferenceImage(reference)
    else:
        resampler.SetSize(reference.size)
        resampler.SetOutputOrigin(reference.origin)
        resampler.SetOutputDirection(reference.direction)
        resampler.SetOutputSpacing(reference.spacing)
    return resampler.Execute(image)


def warp_image(
    image: sitk.Image,
    displacement: sitk.Image,
    reference: Optional[sitk.Image] = None,
    interpolator: int = sitk.sitkLinear,
    padding_value: float = 0,
) -> sitk.Image:
    r"""Interpolate image values at displaced output grid points.

    Args:
        image: Scalar or vector-valued image to evaluate at the specified points.
        displacement: Sampling grid on which to evaluate interpolated continuous image.
        interpolator: Enumeration value of SimpleITK interpolator to use.
        padding_value: Output value when sampling point is outside the input image domain.

    Returns:
        output: Image interpolated at displacement field grid points at the input image
            positions obtained by adding the given displacement to these grid coordinates.

    """
    # cf. https://github.com/SimpleITK/SimpleITK/issues/702
    if reference is None:
        reference = displacement
    resampler = sitk.ResampleImageFilter()
    resampler.SetInterpolator(interpolator)
    resampler.SetDefaultPixelValue(padding_value)
    resampler.SetReferenceImage(reference)
    assert resampler.GetSize() == reference.GetSize()
    # AFTER Use sitk.DisplacementFieldTransform after SetReferenceImage()
    # because it changes disp_field.GetSize() to zero!?!
    disp_field = sitk.Cast(displacement, sitk.sitkVectorFloat64)
    transform = sitk.DisplacementFieldTransform(disp_field)
    resampler.SetTransform(transform)
    return resampler.Execute(image)

    # return sitk.Warp(
    #     image=image,
    #     displacementField=displacement,
    #     interpolator=interpolator,
    #     outputSize=displacement.GetSize(),
    #     outputOrigin=displacement.GetOrigin(),
    #     outputSpacing=displacement.GetSpacing(),
    #     outputDirection=displacement.GetDirection(),
    #     edgePaddingValue=padding_value,
    # )


def interpolate_ndimage(
    image: sitk.Image, points: np.ndarray, padding_value: float = 0, order: int = 1
) -> np.ndarray:
    r"""Use ``scipy.ndimage.map_coordinates`` to interpolate image."""
    grid = image_grid_attributes(image)
    idxs = np.moveaxis(grid.physical_space_to_continuous_index(points), -1, 0)
    idxs = np.flip(idxs, axis=0)
    vals = sitk.GetArrayViewFromImage(image)
    nval = image.GetNumberOfComponentsPerPixel()
    if nval > 1:
        out = np.stack(
            [
                scipy.ndimage.map_coordinates(vals[..., c], idxs, cval=padding_value, order=order)
                for c in range(nval)
            ],
            axis=-1,
        )
    else:
        out = scipy.ndimage.map_coordinates(vals, idxs, cval=padding_value, order=order)
    return out


def interpolate_regular_grid(
    image: sitk.Image, points: np.ndarray, padding_value: float = 0
) -> np.ndarray:
    r"""Use ``scipy.interpolate.RegularGridInterpolator`` to interpolate image data."""
    size = points.shape[0:-1]
    grid = image_grid_attributes(image)
    vals = sitk.GetArrayViewFromImage(image)
    nval = image.GetNumberOfComponentsPerPixel()
    idxs = grid.physical_space_to_continuous_index(points.reshape(-1, grid.ndim))
    idxs = np.flip(idxs, axis=-1)
    coords = tuple(np.arange(vals.shape[axis]) for axis in range(grid.ndim))
    if nval > 1:
        out = []
        for c in range(nval):
            func = scipy.interpolate.RegularGridInterpolator(
                coords, vals[..., c], bounds_error=False, fill_value=padding_value
            )
            out.append(func(idxs).reshape(size).astype(vals.dtype))
        out = np.stack(out, axis=-1)
    else:
        func = scipy.interpolate.RegularGridInterpolator(
            coords, vals, bounds_error=False, fill_value=padding_value
        )
        out = func(idxs).reshape(size).astype(vals.dtype)
    return out


def interpolate_griddata(image: sitk.Image, points: np.ndarray) -> np.ndarray:
    r"""Interpolate image similar to ``scipy.interpolate.griddata``.

    This method should only be used for comparison. The used Delaunay triangulation
    is not suited for interpolating image data sampled on a regular grid.

    Use ``warp_image`` (for regularly spaced ``points``) or ``interpolate_image``.
    """
    size = points.shape[0:-1]
    grid = image_grid_attributes(image)
    tess = scipy.spatial.qhull.Delaunay(grid.points.reshape(-1, grid.ndim))
    vals = sitk.GetArrayViewFromImage(image)
    nval = image.GetNumberOfComponentsPerPixel()
    coor = points.reshape(-1, grid.ndim)
    if nval > 1:
        out = []
        for c in range(nval):
            img = scipy.interpolate.LinearNDInterpolator(tess, vals[..., c].flatten())
            out.append(img(coor).reshape(size))
        out = np.stack(out, axis=-1)
    else:
        img = scipy.interpolate.LinearNDInterpolator(tess, vals.flatten())
        out = img(coor).reshape(size)
    return out



r"""Spatial transformer network (STN)."""

from __future__ import annotations

from copy import deepcopy
from collections import OrderedDict
from dataclasses import dataclass
import math
from typing import Dict, List, Optional, Sequence, Tuple, Union, overload

import torch
from torch import Tensor, Size
from torch.nn import Module, ModuleList, ReLU, Sequential, Upsample

from deepali.core import Device, PathStr, ScalarOrTuple, functional as U
from deepali.core.pathlib import unlink_or_mkdir
from deepali.networks.layers import Conv2d, Conv3d, ConvLayer, Linear, convolution, pooling
from deepali.networks.utils import module_output_size
from deepali.spatial import TransformConfig, has_affine_component, has_nonrigid_component


@dataclass
class SpatialTransformerConfig(TransformConfig):
    r"""Hyperparameters of spatial transformer network (STN)."""

    # Whether to use batch normalization.
    batch_norm_enabled: bool = False
    # Whether to use learnable bias of convolutional layers.
    bias_enabled: bool = True
    # How to initialize biases of convolutational layers
    bias_init: str = "uniform"

    # Maximum rotation angle with respect to unit cube domain.
    affine_max_angle: float = math.pi / 4
    # Maximum translation with respect to unit cube domain.
    affine_max_offset: float = 1
    # Maximum scaling factor with respect to unit cube domain.
    affine_max_scale: float = 1
    # Number of hidden nodes in fully connected hidden layer of affine decoder
    affine_nodes: int = 32
    # Initial parameters of affine output layers
    affine_init: str = "zeros"

    # Whether to divide predicted vector field by image size.
    #
    # This avoids the following issues:
    # 1) An initial translation due to randomly initialized biases having a large impact.
    # 2) STN predictions with very small magnitude, which worsens ITN performance.
    normalize_flow: bool = True

    # List with indices of channels passed from ITN output to STN
    stn_input_channels: list = None

    @staticmethod
    def section() -> str:
        r"""Common key prefix of configuration entries in configuration file."""
        return "model.stn"


class SpatialTransformEncoder(Module):
    r"""Image encoder of spatial transformer network."""

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        bias_enabled: bool = True,
        bias_init: str = "uniform",
    ) -> None:
        super().__init__()

        assert 2 <= spatial_dims <= 3
        acti = ReLU(inplace=True)
        ks = 3

        def conv(m: int, n: int, **kwargs) -> Module:
            bias = bias_init if bias_enabled else False
            pad = ks // 2
            return convolution(spatial_dims, m, n, kernel_size=ks, padding=pad, bias=bias, **kwargs)

        blocks = [
            [
                ("conv_0", conv(in_channels, in_channels, stride=2)),
                ("conv_1", conv(in_channels, 16)),
                ("acti_0", acti),
                ("conv_2", conv(16, 16)),
                ("acti_1", acti),
            ],
            [
                ("conv_0", conv(16, 16, stride=2)),
                ("conv_1", conv(16, 16)),
                ("acti", acti),
            ],
            [
                ("conv_0", conv(16, 32, stride=2)),
                ("conv_1", conv(32, 32)),
                ("acti", acti),
            ],
            [
                ("conv_0", conv(32, 32, stride=2)),
                ("conv_1", conv(32, 32)),
                ("acti", acti),
            ],
        ]
        self.blocks = ModuleList([Sequential(OrderedDict(block)) for block in blocks])

    @property
    def nblocks(self) -> int:
        return len(self.blocks)

    @property
    def spatial_dims(self) -> int:
        conv = self.blocks[0][0]
        if isinstance(conv, Conv2d):
            return 2
        elif isinstance(conv, Conv3d):
            return 3
        raise AssertionError(
            f"{type(self).__name__}.spatial_dims expected first convolution to be Conv2d or Conv3d"
        )

    @property
    def in_channels(self) -> int:
        r"""Number of expected input tensor channels."""
        return self.input_channels()

    @property
    def out_channels(self) -> int:
        r"""Number of channels of last output feature map."""
        return self.output_channels(-1)

    def input_channels(self) -> int:
        r"""Number of input channels."""
        conv = self.blocks[0][0]
        assert isinstance(conv, (Conv2d, Conv3d))
        return conv.weight.shape[1]

    @overload
    def output_channels(self) -> List[int]:
        r"""Number of channels of output feature maps."""
        ...

    @overload
    def output_channels(self, index: int) -> int:
        r"""Number of channels of i-th output feature map."""
        ...

    def output_channels(self, index: Optional[int] = None) -> Union[int, List[int]]:
        r"""Number of channels of output feature map(s)."""
        if index is None:
            channels = []
            for block in self.blocks:
                conv = block[-2]
                assert isinstance(conv, (Conv2d, Conv3d))
                channels.append(conv.weight.shape[0])
            return channels
        nblocks = self.nblocks
        if index < -nblocks or index >= nblocks:
            raise ValueError(
                f"{type(self).__name__}.output_channels() 'index' must be in [-{nblocks}, {nblocks - 1}]"
            )
        if index < 0:
            index += nblocks
        conv = self.blocks[index][-2]
        assert isinstance(conv, (Conv2d, Conv3d))
        return conv.weight.shape[0]

    def output_size(self, in_size: ScalarOrTuple[int], index: int = -1) -> ScalarOrTuple[int]:
        r"""Spatial size of i-th output feature map."""
        nblocks = self.nblocks
        if index < -nblocks or index >= nblocks:
            raise ValueError(
                f"{type(self).__name__}.output_size() 'index' must be in [-{nblocks}, {nblocks - 1}]"
            )
        if index < 0:
            index += nblocks
        ds_factor = 2 ** (index + 1)
        if isinstance(in_size, int):
            return in_size // ds_factor
        return Size((n // ds_factor for n in in_size))

    def output_sizes(self, in_size: ScalarOrTuple[int]) -> List[ScalarOrTuple[int]]:
        r"""Spatial size of output feature maps."""
        sizes = []
        ds_factor = 1
        for _ in self.blocks:
            ds_factor *= 2
            if isinstance(in_size, int):
                sizes.append(in_size // ds_factor)
            else:
                sizes.append(Size((n // ds_factor for n in in_size)))
        return sizes

    def output_shape(self, in_shape: Sequence[int], index: int = -1) -> Size:
        r"""Shape of i-th output feature map."""
        in_channels = self.input_channels()
        if len(in_shape) != self.spatial_dims + 2 or in_shape[1] != in_channels:
            raise ValueError(
                f"{type(self).__name__}.output_shape() 'in_shape' must be of the form"
                f" (N, {in_channels}, {'Y, X' if self.spatial_dims == 2 else 'Z, Y, X'})"
            )
        nblocks = self.nblocks
        if index < -nblocks or index >= nblocks:
            raise ValueError(
                f"{type(self).__name__}.output_shape() 'index' must be in [-{nblocks}, {nblocks - 1}]"
            )
        if index < 0:
            index += nblocks
        batch_size = in_shape[0]
        out_channels = self.output_channels(index)
        out_shape = Size(reversed(self.output_size(in_shape[:1:-1])))
        out_shape = Size((batch_size, out_channels)) + out_shape
        return out_shape

    def forward(self, x: Tensor) -> Tuple[Tensor, ...]:
        out = []
        for block in self.blocks:
            x = block(x)
            out.append(x)
        return tuple(out)


class AffineTransformDecoder(Module):
    r"""Decoder of affine transformation parameters."""

    def __init__(
        self,
        spatial_dims: int,
        features: int,
        hidden: int = 32,
        max_angle: float = math.pi / 4,
        max_offset: float = 1,
        max_scale: float = 1,
        as_matrix: bool = False,
        init: Optional[str] = "zeros",
    ) -> None:
        r"""Initialize affine transformation decoder.

        Args:
            spatial_dims: Number of spatial dimensions.
            features: Number of input features.
            hidden: Number of hidden units.
            max_angle: Maximum rotation angle with respect to unit cube domain.
            max_offset: Maximum translation with respect to unit cube domain.
            max_scale: Maximum scaling factor with respect to unit cube domain.
            as_matrix: Whether to compose and return transformation matrix.

        """
        super().__init__()
        if spatial_dims < 2 or spatial_dims > 3:
            raise ValueError(f"{type(self).__name__}() 'ndim' must be 2 or 3")
        self.spatial_dims = spatial_dims
        self.linear = Sequential(Linear(features, hidden), ReLU(inplace=True))
        self.offset = Linear(hidden, spatial_dims, bias=init, init=init)
        self.angles = Linear(hidden, 1 if spatial_dims == 2 else spatial_dims, bias=init, init=init)
        self.scales = Linear(hidden, spatial_dims, bias=init, init=init)
        self.max_angle = max_angle
        self.max_offset = max_offset
        self.max_scale_log = math.log(max_scale)
        self.as_matrix = as_matrix

    def forward(self, x: Tensor) -> Dict[str, Tensor]:
        r"""Predict affine transformation parameters from encoded features.

        Args:
            x: Input feature vector as tensor of shape ``(N, M)``.

        Returns:
            When ``self.as_matrix == True``, dictionary with key 'affine' with value
            corresponding to the homogeneous coordinate transformation matrix as tensor
            of shape ``(N, D, D + 1)`` where ``D == self.spatial_dims``. Otherwise, dictionary
            keys 'angles', 'offset', and 'scales' contain the separately predicted affine
            transformation parameters.

        """
        x = self.linear(x)
        angles: Tensor = self.angles(x)
        offset: Tensor = self.offset(x)
        scales: Tensor = self.scales(x)
        angles = angles.tanh().mul(self.max_angle)
        offset = offset.tanh().mul(self.max_offset)
        scales = scales.tanh().mul(self.max_scale_log).exp()
        if not self.as_matrix:
            return dict(angles=angles, offset=offset, scales=scales)
        rotation = U.euler_rotation_matrix(angles, order="ZXZ")
        scaling = U.scaling_transform(scales)
        translation = U.translation(offset)
        matrix = U.hmm(scaling, U.hmm(rotation, translation))
        return dict(affine=matrix)


class VectorFieldDecoder(Module):
    r"""Decoder of non-rigid vector field."""

    def __init__(
        self,
        in_channels: int,
        in_size: Sequence[int],
        batch_norm_enabled: bool = False,
        bias_enabled: bool = True,
        bias_init: str = "uniform",
        normalize_flow: bool = True,
    ) -> None:
        if in_channels < 0:
            raise ValueError(f"{type(self).__name__}() 'in_channels' must be positive")
        spatial_dims = len(in_size)
        if spatial_dims < 1 or spatial_dims > 3:
            raise ValueError(f"{type(self).__name__}() 'in_size' must be sequence of length 2 or 3")

        super().__init__()

        self._in_channels = in_channels
        self._in_size = Size(in_size)
        self.normalize_flow = normalize_flow
        interp_mode = "bilinear" if spatial_dims == 2 else "trilinear"

        def conv(m: int, n: int) -> ConvLayer:
            ks = 3
            bias = bias_init if bias_enabled else False
            norm = "batch" if batch_norm_enabled else None
            acti = "relu"
            return ConvLayer(spatial_dims, m, n, kernel_size=ks, bias=bias, norm=norm, acti=acti)

        self.blocks = ModuleList(
            [
                conv(2 * in_channels, 32),
                conv(48, 16),
                conv(32, 16),
                conv(16, 8),
            ]
        )
        self.head = convolution(spatial_dims, 8, spatial_dims, kernel_size=3, padding=1)
        self.up = Upsample(scale_factor=2, mode=interp_mode, align_corners=False)

    @property
    def nblocks(self) -> int:
        return len(self.blocks)

    @property
    def spatial_dims(self) -> int:
        return len(self.in_size)

    @property
    def in_channels(self) -> int:
        return self._in_channels

    @property
    def in_size(self) -> Size:
        return self._in_size

    @property
    def out_channels(self) -> int:
        return self.spatial_dims

    @property
    def out_size(self) -> Size:
        r"""Spatial size of output vector field parameters."""
        return self.output_size()

    @property
    def out_shape(self) -> Size:
        r"""Shape of output vector field parameters for a batch size of 1."""
        return self.output_shape()

    def input_shape(self, batch_size: int = 1) -> Size:
        r"""Expected shape of input tensor for given batch size."""
        return Size((batch_size, self.in_channels)) + self.in_size[::-1]

    def output_shape(self, in_shape: Optional[Sequence[int]] = None) -> Size:
        r"""Calculate shape of output vector field parameters."""
        if not in_shape:
            batch_size = 1
            in_shape = self.input_shape(batch_size)
        else:
            batch_size = in_shape[0]
            input_shape = self.input_shape(batch_size)
            if len(in_shape) != self.spatial_dims + 2 or in_shape != input_shape:
                raise ValueError(
                    f"{type(self).__name__}.output_shape() 'in_shape' must be"
                    f" (N, {', '.join((str(n) for n in input_shape[1:]))}"
                )
        return Size((batch_size, self.out_channels)) + self.output_size(in_shape[:1:-1])[::-1]

    def output_size(self, in_size: Optional[ScalarOrTuple[int]] = None) -> ScalarOrTuple[int]:
        r"""Calculate spatial size of vector field parameters."""
        if in_size is not None and in_size != self.in_size:
            raise ValueError(
                f"{type(self).__name__}.output_size() 'in_size' must be None or {self.in_size}"
            )
        size = self.in_size
        for i, block in enumerate(self.blocks):
            if i < len(self.blocks) - 1:
                size = module_output_size(self.up, size)
            size = module_output_size(block, size)
        size = module_output_size(self.head, size)
        return size

    def forward(self, features: Sequence[Tensor]) -> Tensor:
        features: List[Tensor] = list(features)
        if len(features) != len(self.blocks):
            raise ValueError(
                f"{type(self).__name__} expected {len(self.nblocks)}"
                f" input feature map{'s' if len(self.nblocks) > 1 else ''}"
            )
        x = features.pop()
        for block in self.blocks:
            if features:
                x = torch.cat([self.up(x), features.pop()], dim=1)
            x = block(x)
        x = self.head(x)
        # Avoids the following issues:
        # 1) an initial translation due to randomly initialized biases having a large impact,
        # 2) STN predictions with very small magnitude, which worsens ITN performance
        if self.normalize_flow:
            x = U.normalize_flow(x, size=self.in_size, side_length=1, align_corners=True)
        return x


class SpatialTransformerNetwork(Module):
    r"""Spatial transformer network.

    The input to the STN is generally a concatenated pair of images with one or more channels each,
    and the output is a vector field. The predicted vectors can be either displacement vectors, velocity
    vectors, or control point coefficients of a spline representation. The interpretation of the output
    vectors depends on their subsequent use as parameters of a ``SpatialTransform``, which represents
    the actual spatial coordinate transformation used to deform an image grid or other set of points
    at which an input moving image is sampled to produce a deformed output image.

    """

    def __init__(
        self, in_channels: int, in_size: Sequence[int], config: SpatialTransformerConfig
    ) -> None:
        r"""Initialize spatial transformer modules.

        Args:
            in_channels: Number of input tensor channels.
            in_size: Number and size of spatial dimensions of input tensor.
            config: Spatial transformer network configuration.

        """
        super().__init__()
        config = deepcopy(config)
        self.config = config
        self._in_channels = in_channels
        self._in_size = Size(in_size)
        spatial_dims = len(in_size)
        # Encoder of concatenated input images
        self.encoder = SpatialTransformEncoder(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            bias_enabled=config.bias_enabled,
            bias_init=config.bias_init,
        )
        # Number and spatial size of feature maps at coarsest scale
        fm_channels = self.encoder.output_channels(-1)
        fm_size = self.encoder.output_size(in_size)
        # Decoders of output transformation parameters
        if has_affine_component(config.transform):
            ds_factor = 2
            features = fm_channels * Size((n // ds_factor for n in fm_size)).numel()
            self.pool = pooling("avg", spatial_dims=spatial_dims, kernel_size=ds_factor)
            self.affine = AffineTransformDecoder(
                spatial_dims=spatial_dims,
                features=features,
                hidden=config.affine_nodes,
                max_angle=config.affine_max_angle,
                max_offset=config.affine_max_offset,
                max_scale=config.affine_max_scale,
                as_matrix=config.affine_model == "A",
                init=config.affine_init,
            )
        else:
            self.affine = None
        if has_nonrigid_component(config.transform):
            self.vfield = VectorFieldDecoder(
                in_channels=fm_channels,
                in_size=fm_size[::-1] if config.flip_grid_coords else fm_size,
                bias_enabled=config.bias_enabled,
                bias_init=config.bias_init,
                batch_norm_enabled=config.batch_norm_enabled,
                normalize_flow=config.normalize_flow,
            )
        else:
            self.vfield = None

    @property
    def spatial_dims(self) -> int:
        r"""Number of spatial dimensions."""
        return len(self.in_size)

    @property
    def in_channels(self) -> int:
        return self._in_channels

    @property
    def in_size(self) -> Size:
        return self._in_size

    @property
    def in_shape(self) -> Size:
        r"""Expected input tensor shape for a batch size of 1."""
        return self.input_shape()

    def input_shape(self, batch_size: int = 1) -> Size:
        r"""Expected input tensor shape for given batch size."""
        return Size((batch_size, self.in_channels)) + self.in_size[::-1]

    @property
    def out_channels(self) -> int:
        r"""Number of output channels of vector field parameters or 0."""
        return 0 if self.vfield is None else self.spatial_dims

    @property
    def out_size(self) -> Size:
        r"""Spatial size of output vector field parameters for batch size 1 or empty Size."""
        return self.output_size()

    @property
    def out_shape(self) -> Size:
        r"""Shape of output vector field parameters for batch size 1 or empty Size."""
        return self.output_shape()

    def output_shape(self, in_shape: Optional[Sequence[int]] = None) -> Size:
        r"""Calculate shape of output vector field parameters."""
        if not in_shape:
            batch_size = 1
        else:
            batch_size = in_shape[0]
            input_shape = self.input_shape(batch_size)
            if len(in_shape) != self.spatial_dims + 2 or in_shape != input_shape:
                raise ValueError(
                    f"{type(self).__name__}.output_shape() 'in_shape' must be"
                    f" (N, {', '.join((str(n) for n in input_shape[1:]))}"
                )
        size = self.output_size()
        if not size:
            return Size()
        return Size((batch_size, self.spatial_dims)) + size[::-1]

    def output_size(self, in_size: Optional[ScalarOrTuple[int]] = None) -> ScalarOrTuple[int]:
        r"""Calculate spatial size of output vector field parameters."""
        if in_size and in_size != self.in_size:
            raise ValueError(
                f"{type(self).__name__}.output_size() 'in_size' must be None or {self.in_size}"
            )
        if self.vfield is None:
            return Size()
        size = self.encoder.output_size(self.in_size)
        size = self.vfield.output_size(size)
        return size

    def forward(self, x: Tensor) -> Dict[str, Tensor]:
        r"""Infer spatial transformation parameters given input data tensor."""
        params = {}
        features: Tensor = self.encoder(x)
        # Affine parameters
        if self.affine is not None:
            z: Tensor = features[-1]
            if self.pool is not None:
                z = self.pool(z)
            params.update(self.affine(z.flatten(1)))
        # Non-rigid parameters
        if self.vfield is not None:
            params["vfield"] = self.vfield(features)
        return params

    @classmethod
    def read(
        cls,
        path: PathStr,
        in_channels: int,
        in_size: Sequence[int],
        device: Optional[Device] = None,
    ) -> SpatialTransformerNetwork:
        r"""Create spatial transformer network from previously saved model file.

        Args:
            path: Path of saved model file written by ``self.write(path)``.
            in_channels: Number of input channels.
            in_size: Spatial size of input tensor.
            device: Device on which to load model. If ``None``, use "cpu".

        Returns:
            Spatial transformer network with parameters initialized to the previously saved state.

        """
        if device is None:
            device = torch.device("cpu")
        data = torch.load(path, map_location=device)
        if not isinstance(data, dict):
            raise ValueError(f"{cls.__name__}.read() model file must contain dictionary")
        if "config" not in data:
            raise ValueError(f"{cls.__name__}.read() model file must contain 'config' dict entry")
        if "state" not in data:
            raise ValueError(f"{cls.__name__}.read() model file must contain 'state' dict entry")
        config = SpatialTransformerConfig.from_dict(data["config"])
        stn = cls(in_channels=in_channels, in_size=in_size, config=config)
        stn.to(device).load_state_dict(data["state"])
        return stn

    def write(self, path: PathStr) -> None:
        r"""Save model parameters and configuration."""
        path = unlink_or_mkdir(path)
        torch.save({"config": self.config.asdict(), "state": self.state_dict()}, path)



r"""Convolutional layers."""

# TODO: ConvLayer() support use of dropout, by default between N and A.

from collections import OrderedDict
import math
from numbers import Integral
from typing import Any, Optional, Union

from torch.nn import Module, init
from torch.nn.modules import Sequential
from torch.nn.modules import Conv1d as _Conv1d, ConvTranspose1d as _ConvTranspose1d
from torch.nn.modules import Conv2d as _Conv2d, ConvTranspose2d as _ConvTranspose2d
from torch.nn.modules import Conv3d as _Conv3d, ConvTranspose3d as _ConvTranspose3d

from deepali.core.enum import PaddingMode
from deepali.core.nnutils import same_padding, stride_minus_kernel_padding
from deepali.core.typing import ScalarOrTuple, ScalarOrTuple1d, ScalarOrTuple2d, ScalarOrTuple3d
from deepali.modules import ReprWithCrossReferences

from .acti import ActivationArg, activation
from .norm import NormArg, normalization


__all__ = (
    "Conv1d",
    "Conv2d",
    "Conv3d",
    "ConvTranspose1d",
    "ConvTranspose2d",
    "ConvTranspose3d",
    "ConvLayer",
    "convolution",
    "conv_module",
    "is_convolution",
    "is_conv_module",
    "is_transposed_convolution",
    "same_padding",
    "stride_minus_kernel_padding",
)


class _ConvInit(object):
    r"""Mix-in for initialization of convolutional layer parameters."""

    def reset_parameters(self) -> None:
        # Initialize weights
        if self.weight_init in ("default", "uniform"):
            init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        elif self.weight_init == "xavier":
            init.xavier_uniform_(self.weight)
        elif self.weight_init == "constant":
            init.constant_(self.weight, 0.1)
        elif self.weight_init == "zeros":
            init.constant_(self.weight, 0.0)
        else:
            raise AssertionError(
                f"{type(self).__name__}.reset_parameters() invalid 'init' value: {self.weight_init!r}"
            )
        # Initialize bias
        if self.bias is not None:
            if self.bias_init in ("default", "uniform"):
                fan_in, _ = init._calculate_fan_in_and_fan_out(self.weight)
                bound = 1 / math.sqrt(fan_in)
                init.uniform_(self.bias, -bound, bound)
            elif self.bias_init == "constant":
                init.constant_(self.bias, 0.1)
            elif self.bias_init == "zeros":
                init.constant_(self.bias, 0.0)
            else:
                raise AssertionError(
                    f"{type(self).__name__}.reset_parameters() invalid 'bias' value: {self.bias_init!r}"
                )


class Conv1d(_ConvInit, _Conv1d):
    r"""Convolutional layer with custom initialization of learnable parameters."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple1d[int],
        stride: ScalarOrTuple1d[int] = 1,
        padding: ScalarOrTuple1d[int] = 0,
        dilation: ScalarOrTuple1d[int] = 1,
        groups: int = 1,
        bias: Union[bool, str] = True,
        padding_mode: str = "zeros",
        init: str = "default",
    ):
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = "uniform" if init == "default" else init
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bool(bias),
            padding_mode=padding_mode,
        )


class Conv2d(_ConvInit, _Conv2d):
    r"""Convolutional layer with custom initialization of learnable parameters."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple2d[int],
        stride: ScalarOrTuple2d[int] = 1,
        padding: ScalarOrTuple2d[int] = 0,
        dilation: ScalarOrTuple2d[int] = 1,
        groups: int = 1,
        bias: Union[bool, str] = True,
        padding_mode: str = "zeros",
        init: str = "default",
    ):
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = "uniform" if init == "default" else init
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bool(bias),
            padding_mode=padding_mode,
        )


class Conv3d(_ConvInit, _Conv3d):
    r"""Convolutional layer with custom initialization of learnable parameters."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple3d[int],
        stride: ScalarOrTuple3d[int] = 1,
        padding: ScalarOrTuple3d[int] = 0,
        dilation: ScalarOrTuple3d[int] = 1,
        groups: int = 1,
        bias: Union[bool, str] = True,
        padding_mode: str = "zeros",
        init: str = "uniform",
    ):
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = "uniform" if init == "default" else init
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bool(bias),
            padding_mode=padding_mode,
        )


class ConvTranspose1d(_ConvInit, _ConvTranspose1d):
    r"""Transposed convolution in 1D with custom initialization of learnable parameters."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple1d[int],
        stride: ScalarOrTuple1d[int] = 1,
        padding: ScalarOrTuple1d[int] = 0,
        output_padding: ScalarOrTuple1d[int] = 0,
        dilation: ScalarOrTuple1d[int] = 1,
        groups: int = 1,
        bias: Union[bool, str] = True,
        padding_mode: str = "zeros",
        init: str = "default",
    ):
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = "uniform" if init == "default" else init
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            bias=bool(bias),
            padding_mode=padding_mode,
        )


class ConvTranspose2d(_ConvInit, _ConvTranspose2d):
    r"""Transposed convolution in 2D with custom initialization of learnable parameters."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple2d[int],
        stride: ScalarOrTuple2d[int] = 1,
        padding: ScalarOrTuple2d[int] = 0,
        output_padding: ScalarOrTuple2d[int] = 0,
        dilation: ScalarOrTuple2d[int] = 1,
        groups: int = 1,
        bias: Union[bool, str] = True,
        padding_mode: str = "zeros",
        init: str = "default",
    ):
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = "uniform" if init == "default" else init
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            bias=bool(bias),
            padding_mode=padding_mode,
        )


class ConvTranspose3d(_ConvInit, _ConvTranspose3d):
    r"""Transposed convolution in 3D with custom initialization of learnable parameters."""

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple3d[int],
        stride: ScalarOrTuple3d[int] = 1,
        padding: ScalarOrTuple3d[int] = 0,
        output_padding: ScalarOrTuple3d[int] = 0,
        dilation: ScalarOrTuple3d[int] = 1,
        groups: int = 1,
        bias: Union[bool, str] = True,
        padding_mode: str = "zeros",
        init: str = "default",
    ):
        self.bias_init = "uniform" if isinstance(bias, bool) else bias
        self.weight_init = "uniform" if init == "default" else init
        super().__init__(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            output_padding=output_padding,
            dilation=dilation,
            groups=groups,
            bias=bool(bias),
            padding_mode=padding_mode,
        )


class ConvLayer(ReprWithCrossReferences, Sequential):
    r"""Convolutional layer with optional pre- or post-convolution normalization and/or activation."""

    def __init__(
        self,
        spatial_dims: int,
        in_channels: int,
        out_channels: int,
        kernel_size: ScalarOrTuple[int],
        stride: ScalarOrTuple[int] = 1,
        padding: Optional[ScalarOrTuple[int]] = None,
        output_padding: Optional[ScalarOrTuple[int]] = None,
        padding_mode: Union[PaddingMode, str] = "zeros",
        dilation: ScalarOrTuple[int] = 1,
        groups: int = 1,
        init: str = "default",
        bias: Optional[Union[bool, str]] = None,
        norm: NormArg = None,
        acti: ActivationArg = None,
        order: str = "CNA",
        transposed: bool = False,
        conv: Optional[Module] = None,
    ) -> None:
        if spatial_dims < 0 or spatial_dims > 3:
            raise ValueError("ConvLayer() 'spatial_dims' must be 1, 2, or 3")
        if isinstance(kernel_size, Integral):
            kernel_size = (int(kernel_size),) * spatial_dims
        if padding is None:
            padding = same_padding(kernel_size, dilation)
        # Order of layer operations
        if not isinstance(order, str):
            raise TypeError("ConvLayer() 'order' must be str")
        order = order.upper()
        if "C" not in order:
            raise ValueError("ConvLayer() 'order' must contain 'C' for convolution")
        if "D" in order:
            raise NotImplementedError("ConvLayer() 'order' has 'D' for not implemented dropout")
        if len(set(order)) != len(order) or any(c not in {"A", "C", "N"} for c in order):
            raise ValueError(
                "ConvLayer() 'order' must be permutation of characters"
                f" 'A' (activation), 'N' (norm), and 'C' (convolution), got {order!r}"
            )
        # Non-linear activation
        if acti is None or "A" not in order:
            acti_fn = None
        else:
            acti_fn = activation(acti)
        # Normalization layer
        norm_after_conv = False
        if norm is None or "N" not in order:
            norm_layer = None
        else:
            norm_after_conv = order.index("C") < order.index("N")
            num_features = out_channels if norm_after_conv else in_channels
            norm_layer = normalization(norm, spatial_dims=spatial_dims, num_features=num_features)
        # Convolution layer
        if bias is None:
            bias = True
            if norm_after_conv and isinstance(norm_layer, Module):
                bias = not getattr(norm_layer, "affine", False)
        if output_padding is None:
            output_padding = stride_minus_kernel_padding(1, stride)
        if conv is None:
            conv = convolution(
                spatial_dims,
                in_channels,
                out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                output_padding=output_padding,
                padding_mode=padding_mode,
                dilation=dilation,
                groups=groups,
                bias=bias,
                init=init,
                transposed=transposed,
            )
        # Initialize module
        modules = {
            "A": ("acti", acti_fn),
            "N": ("norm", norm_layer),
            "C": ("conv", conv),
        }
        modules = OrderedDict(modules[k] for k in order if modules[k][1] is not None)
        super().__init__(modules)
        if acti_fn is None:
            self.acti = None
        if norm_layer is None:
            self.norm = None
        self.spatial_dims = spatial_dims
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.order = order

    def has_norm_after_conv(self) -> bool:
        r"""Whether this layer has a normalization layer after the convolution."""
        names = tuple(name for name, _ in self.named_modules() if name in ("conv", "norm"))
        return len(names) == 2 and names[-1] == "norm"


def convolution(
    spatial_dims: int,
    in_channels: int,
    out_channels: int,
    kernel_size: ScalarOrTuple[int],
    stride: ScalarOrTuple[int] = 1,
    padding: Optional[ScalarOrTuple[int]] = 0,
    output_padding: Optional[ScalarOrTuple[int]] = 0,
    padding_mode: Union[PaddingMode, str] = "zeros",
    dilation: ScalarOrTuple[int] = 1,
    groups: int = 1,
    init: str = "default",
    bias: Union[bool, str] = True,
    transposed: bool = False,
) -> Module:
    r"""Create convolution module for specified number of spatial input tensor dimensions."""
    if in_channels < 1:
        raise ValueError(f"convolution() 'in_channels' ({in_channels}) must be positive")
    if out_channels < 1:
        raise ValueError(f"convolution() 'out_channels' ({out_channels}) must be positive")
    padding_mode = PaddingMode.from_arg(padding_mode)
    if padding_mode is PaddingMode.NONE:
        padding = 0
    elif padding is None:
        padding = same_padding(kernel_size, dilation)
    kwargs = dict(
        kernel_size=kernel_size,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
        bias=bias,
        init=init,
    )
    # Specify non-default padding_mode only if used so conv.extra_repr() does not show it
    if any(n != 0 for n in ((padding,) if isinstance(padding, int) else padding)):
        kwargs["padding_mode"] = padding_mode.conv_mode(spatial_dims)
    if transposed:
        if output_padding is None:
            output_padding = stride_minus_kernel_padding(1, stride)
        kwargs["output_padding"] = output_padding
    if spatial_dims == 1:
        conv_type = ConvTranspose1d if transposed else Conv1d
    elif spatial_dims == 2:
        conv_type = ConvTranspose2d if transposed else Conv2d
    elif spatial_dims == 3:
        conv_type = ConvTranspose3d if transposed else Conv3d
    else:
        raise ValueError("convolution() 'spatial_dims' must be 1, 2, or 3")
    return conv_type(in_channels, out_channels, **kwargs)


def conv_module(*args, **kwargs) -> Module:
    r"""Create convolution layer, see ``convolution()``."""
    return convolution(*args, **kwargs)


def is_convolution(arg: Any) -> bool:
    r"""Whether given module is a learnable convolution."""
    types = (
        _Conv1d,
        _Conv2d,
        _Conv3d,
        _ConvTranspose1d,
        _ConvTranspose2d,
        _ConvTranspose3d,
    )
    return isinstance(arg, types)


def is_conv_module(arg: Any) -> bool:
    r"""Whether given module is a learnable convolution."""
    return is_convolution(arg)


def is_transposed_convolution(arg: Any) -> bool:
    r"""Whether given module is a learnable transposed convolution."""
    types = (_ConvTranspose1d, _ConvTranspose2d, _ConvTranspose3d)
    return isinstance(arg, types)



r"""Metrics for multi-label classification performance evaluation."""

from typing import Callable, Optional, Sequence, Union

from ignite.engine import Engine
from ignite.exceptions import NotComputableError
from ignite.metrics.metric import Metric, reinit__is_reduced, sync_all_reduce

import torch
from torch import Tensor


class MultiLabelScore(Metric):
    r"""Compute a score for each class label."""

    def __init__(
        self,
        score_fn: Callable,
        num_classes: int,
        output_transform: Callable = lambda x: x,
        device: Optional[Union[str, torch.device]] = None,
    ):
        self.score_fn = score_fn
        self.num_classes = num_classes
        self.accumulator = None
        self.num_examples = 0
        super().__init__(output_transform=output_transform, device=device)

    @reinit__is_reduced
    def reset(self) -> None:
        self.accumulator = torch.zeros(self.num_classes, dtype=torch.float32, device=self._device)
        self.num_examples = 0

    @reinit__is_reduced
    def update(self, output: Sequence[Tensor]) -> None:
        y_pred, y = output
        if y_pred.ndim < 2:
            raise ValueError(
                "MultiLabelScore.update() y_pred must have shape (N, C, ...),"
                f" but given {y_pred.shape}"
            )
        if y_pred.shape[1] not in (1, self.num_classes):
            raise ValueError(
                f"MultiLabelScore.update() expected y_pred to have 1 or {self.num_channels} channels"
            )
        if y.ndim + 1 == y_pred.ndim:
            y = y.unsqueeze(1)
        elif y.ndim != y_pred.ndim:
            raise ValueError(
                "MultiLabelScore.update() y_pred must have shape (N, C, ...) and y must have"
                f" shape (N, ...) or (N, 1, ...), but given {y.shape} vs {y_pred.shape}"
            )
        if y.shape != (y_pred.shape[0], 1) + y_pred.shape[2:]:
            raise ValueError("y and y_pred must have compatible shapes.")
        scores = multilabel_score(self.score_fn, y_pred, y, num_classes=self.num_classes)
        self.accumulator += scores
        self.num_examples += y_pred.shape[0]

    @sync_all_reduce("accumulator", "num_examples")
    def compute(self) -> float:
        if self.num_examples == 0:
            raise NotComputableError(
                "Loss must have at least one example before it can be computed."
            )
        return self.accumulator / self.num_examples

    @torch.no_grad()
    def iteration_completed(self, engine: Engine) -> None:
        output = self._output_transform(engine.state.output)
        self.update(output)


def multilabel_score(
    score_fn, preds: Tensor, labels: Tensor, num_classes: Optional[int] = None
) -> Tensor:
    r"""Evaluate score for each class label."""
    assert labels.shape[1] == 1
    if num_classes is None:
        num_classes = preds.shape[1]
        if num_classes == 1:
            raise ValueError(
                "multilabel_score() 'num_classes' required when 'preds' is not one-hot encoded"
            )
    if preds.shape[1] == num_classes:
        preds = preds.argmax(dim=1, keepdim=True)
    elif preds.shape[1] != 1:
        raise ValueError("multilabel_score() 'preds' must have shape (N, C|1, ..., X)")
    result = torch.zeros(num_classes, dtype=torch.float, device=preds.device)
    for label in range(num_classes):
        y_pred = preds.eq(label).float()
        y = labels.eq(label).float()
        result[label] = score_fn(y_pred, y).mean()
    return result



r"""Auxiliary functions for Jupyter notebooks."""

from IPython import get_ipython


def is_jupyter_notebook() -> bool:
    r"""Check if code is running in Jupyter notebook."""
    try:
        ipython = get_ipython()
        if ipython is None or "IPKernelApp" not in ipython.config:
            return False
    except ImportError:
        return False
    return True



r"""Utilities for working with temporary files and directories."""

from contextlib import contextmanager
import os
from pathlib import Path
import shutil
from tempfile import mkdtemp, mkstemp
from typing import Generator, Optional, Union

from .pathlib import PathStr, delete


__all__ = (
    "make_temp_file",
    "temp_dir",
    "temp_file",
)


def make_temp_file(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    dir: Optional[PathStr] = None,
    text: bool = False,
) -> Path:
    r"""Make temporary file with mkstemp, but close open file handle immediately."""
    fp, path = mkstemp(suffix=suffix, prefix=prefix, dir=dir, text=text)
    os.close(fp)
    return Path(path)


@contextmanager
def temp_dir(
    suffix: str = None, prefix: str = None, dir: Union[Path, str] = None
) -> Generator[Path, None, None]:
    r"""Create temporary directory within context."""
    path = mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        yield Path(path)
    finally:
        shutil.rmtree(path)


@contextmanager
def temp_file(
    suffix: Optional[str] = None,
    prefix: Optional[str] = None,
    dir: PathStr = None,
    text: bool = False,
) -> Generator[Path, None, None]:
    r"""Create temporary file within context."""
    path = make_temp_file(suffix=suffix, prefix=prefix, dir=dir, text=text)
    try:
        yield path
    finally:
        delete(path)



import pytest

import numpy as np
import torch
from torch import Tensor

from deepali.core import Cube, Grid
from deepali.core import affine as A


@pytest.fixture
def default_angle() -> Tensor:
    return torch.deg2rad(torch.tensor(33.0))


@pytest.fixture
def default_cube(default_angle: Tensor) -> Cube:
    direction = A.euler_rotation_matrix(default_angle)
    cube = Cube(extent=(34, 42), center=(7, 4), direction=direction)
    return cube


def test_cube_is_grid_with_three_points(default_cube: Cube) -> None:
    cube = default_cube
    grid = cube.grid(size=3, align_corners=True)
    assert type(grid) is Grid
    assert grid.ndim == cube.ndim
    assert grid.device == cube.device
    assert grid.size() == (3,) * grid.ndim
    assert grid.align_corners() is True
    assert torch.allclose(grid.cube_extent(), cube.extent())
    assert torch.allclose(grid.affine(), cube.affine())
    assert torch.allclose(grid.inverse_affine(), cube.inverse_affine())
    assert torch.allclose(grid.spacing(), cube.spacing())
    assert torch.allclose(grid.transform("cube_corners", "world"), cube.transform("cube", "world"))
    assert torch.allclose(grid.transform("world", "cube_corners"), cube.transform("world", "cube"))
    assert torch.allclose(grid.transform(), cube.transform())
    assert torch.allclose(grid.inverse_transform(), cube.inverse_transform())


def test_cube_to_from_numpy(default_cube: Cube) -> None:
    r"""Test converting a Cube to a 1-dimensional NumPy array and constructing a new one from such array."""
    dim = default_cube.ndim
    arr = default_cube.numpy()
    assert isinstance(arr, np.ndarray)
    assert arr.ndim == 1
    assert arr.dtype == (np.float32 if default_cube.dtype == torch.float32 else np.float64)
    assert arr.shape == ((dim + 2) * dim,)
    cube = Cube.from_numpy(arr)
    assert cube == default_cube


def test_cube_eq(default_cube: Cube, default_angle: Tensor) -> None:
    r"""Test comparison of different Cube instances for equality."""
    # Same instance
    assert default_cube == default_cube

    # Different instance, same atributes
    other_cube = Cube(
        extent=default_cube.extent(),
        center=default_cube.center(),
        direction=default_cube.direction(),
    )
    assert default_cube == other_cube

    # Different extent
    other_cube = Cube(
        extent=default_cube.extent().add(1),
        center=default_cube.center(),
        direction=default_cube.direction(),
    )
    assert default_cube != other_cube

    # Different center
    other_cube = Cube(
        extent=default_cube.extent(),
        center=default_cube.center().add(0.001),
        direction=default_cube.direction(),
    )
    assert default_cube != other_cube

    # Different direction
    other_direction = A.euler_rotation_matrix(default_angle.add(0.001))
    other_cube = Cube(
        extent=default_cube.extent(),
        center=default_cube.center(),
        direction=other_direction,
    )
    assert default_cube != other_cube



r"""The transforms in this Python package generally built on the :mod:`.core` library.
The classes defined by these modules can be used, for example, in a data input pipeline
which is attached to a data loader. The spatial transforms defined in the :mod:`.spatial`
library, on the other hand, can be used to implement either a traditional or machine
learning based co-registration approach.

Note that data transforms are included in the :mod:`.data` library to avoid cyclical
imports between modules defining specialized tensor types such as :mod:`.data.image`
and datasets defined in :mod:`.data.dataset`, which also use these transforms to read
and preprocess the loaded data.

Following torchvision's lead, data transform classes which operate on tensors and do not require
lambda functions are derived from ``torch.nn.Module``. Use ``torch.nn.Sequential`` to compose
transforms instead of ``torchvision.transforms.Compose``. This is to support ``torch.jit.script``.

See also: https://github.com/pytorch/vision/blob/3852b41975702cb683a92c8e37f1ef74fd6a79b1/torchvision/transforms/transforms.py#L49.

"""

from typing import Callable

from .image import AvgPoolImage
from .image import CastImage
from .image import CenterCropImage
from .image import CenterPadImage
from .image import ClampImage
from .image import ImageToTensor
from .image import NarrowImage
from .image import NormalizeImage
from .image import ReadImage
from .image import ResampleImage
from .image import RescaleImage
from .image import ResizeImage

from .item import ItemTransform
from .item import ItemwiseTransform


Transform = Callable


__all__ = (
    # Type annotation
    "Transform",
    # Generic transforms
    "ItemTransform",
    "ItemwiseTransform",
    # Image transforms
    "AvgPoolImage",
    "CastImage",
    "CenterCropImage",
    "CenterPadImage",
    "ClampImage",
    "ImageToTensor",
    "NarrowImage",
    "NormalizeImage",
    "ReadImage",
    "ResampleImage",
    "RescaleImage",
    "ResizeImage",
)
